import numpy as np
from gym import utils
from gym.envs.mujoco import mujoco_env

from imitation.envs.examples.airl_envs.dynamic_mjc.model_builder import MJCModel


def ant_env(gear=150, eyes=True):
    mjcmodel = MJCModel("ant_maze")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="degree", coordinate="local")
    mjcmodel.root.option(timestep="0.01", integrator="RK4")
    mjcmodel.root.custom().numeric(
        data="0.0 0.0 0.55 1.0 0.0 0.0 0.0 0.0 1.0 0.0 -1.0 0.0 -1.0 0.0 1.0",
        name="init_qpos",
    )
    asset = mjcmodel.root.asset()
    asset.texture(
        builtin="gradient",
        height="100",
        rgb1="1 1 1",
        rgb2="0 0 0",
        type="skybox",
        width="100",
    )
    asset.texture(
        builtin="flat",
        height="1278",
        mark="cross",
        markrgb="1 1 1",
        name="texgeom",
        random="0.01",
        rgb1="0.8 0.6 0.4",
        rgb2="0.8 0.6 0.4",
        type="cube",
        width="127",
    )
    asset.texture(
        builtin="checker",
        height="100",
        name="texplane",
        rgb1="0 0 0",
        rgb2="0.8 0.8 0.8",
        type="2d",
        width="100",
    )
    asset.material(
        name="MatPlane",
        reflectance="0.5",
        shininess="1",
        specular="1",
        texrepeat="60 60",
        texture="texplane",
    )
    asset.material(name="geom", texture="texgeom", texuniform="true")

    default = mjcmodel.root.default()
    default.joint(armature=1, damping=1, limited="true")
    default.geom(
        friction=[1.5, 0.5, 0.5],
        density=5.0,
        margin=0.01,
        condim=3,
        conaffinity=0,
        rgba="0.8 0.6 0.4 1",
    )

    worldbody = mjcmodel.root.worldbody()
    worldbody.light(
        cutoff="100",
        diffuse=[0.8, 0.8, 0.8],
        dir="-0 0 -1.3",
        directional="true",
        exponent="1",
        pos="0 0 1.3",
        specular=".1 .1 .1",
    )
    worldbody.geom(
        conaffinity=1,
        condim=3,
        material="MatPlane",
        name="floor",
        pos="0 0 0",
        rgba="0.8 0.9 0.8 1",
        size="40 40 40",
        type="plane",
    )

    ant = worldbody.body(name="torso", pos=[0, 0, 0.75])
    ant.geom(name="torso_geom", pos=[0, 0, 0], size="0.25", type="sphere")
    ant.joint(
        armature="0",
        damping="0",
        limited="false",
        margin="0.01",
        name="root",
        pos=[0, 0, 0],
        type="free",
    )

    if eyes:
        eye_z = 0.1
        eye_y = -0.21
        eye_x_offset = 0.07
        # eyes
        ant.geom(
            fromto=[eye_x_offset, 0, eye_z, eye_x_offset, eye_y, eye_z],
            name="eye1",
            size="0.03",
            type="capsule",
            rgba=[1, 1, 1, 1],
        )
        ant.geom(
            fromto=[eye_x_offset, 0, eye_z, eye_x_offset, eye_y - 0.02, eye_z],
            name="eye1_",
            size="0.02",
            type="capsule",
            rgba=[0, 0, 0, 1],
        )
        ant.geom(
            fromto=[-eye_x_offset, 0, eye_z, -eye_x_offset, eye_y, eye_z],
            name="eye2",
            size="0.03",
            type="capsule",
            rgba=[1, 1, 1, 1],
        )
        ant.geom(
            fromto=[-eye_x_offset, 0, eye_z, -eye_x_offset, eye_y - 0.02, eye_z],
            name="eye2_",
            size="0.02",
            type="capsule",
            rgba=[0, 0, 0, 1],
        )
        # eyebrows
        ant.geom(
            fromto=[
                eye_x_offset - 0.03,
                eye_y,
                eye_z + 0.07,
                eye_x_offset + 0.03,
                eye_y,
                eye_z + 0.1,
            ],
            name="brow1",
            size="0.02",
            type="capsule",
            rgba=[0, 0, 0, 1],
        )
        ant.geom(
            fromto=[
                -eye_x_offset + 0.03,
                eye_y,
                eye_z + 0.07,
                -eye_x_offset - 0.03,
                eye_y,
                eye_z + 0.1,
            ],
            name="brow2",
            size="0.02",
            type="capsule",
            rgba=[0, 0, 0, 1],
        )

    front_left_leg = ant.body(name="front_left_leg", pos=[0, 0, 0])
    front_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="aux_1_geom",
        size="0.08",
        type="capsule",
    )
    aux_1 = front_left_leg.body(name="aux_1", pos=[0.2, 0.2, 0])
    aux_1.joint(
        axis=[0, 0, 1], name="hip_1", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="left_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_1 = aux_1.body(pos=[0.2, 0.2, 0])
    ankle_1.joint(
        axis=[-1, 1, 0],
        name="ankle_1",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, 0.4, 0.0],
        name="left_ankle_geom",
        size="0.08",
        type="capsule",
    )

    front_right_leg = ant.body(name="front_right_leg", pos=[0, 0, 0])
    front_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="aux_2_geom",
        size="0.08",
        type="capsule",
    )
    aux_2 = front_right_leg.body(name="aux_2", pos=[-0.2, 0.2, 0])
    aux_2.joint(
        axis=[0, 0, 1], name="hip_2", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="right_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_2 = aux_2.body(pos=[-0.2, 0.2, 0])
    ankle_2.joint(
        axis=[1, 1, 0],
        name="ankle_2",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, 0.4, 0.0],
        name="right_ankle_geom",
        size="0.08",
        type="capsule",
    )

    back_left_leg = ant.body(name="back_left_leg", pos=[0, 0, 0])
    back_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="aux_3_geom",
        size="0.08",
        type="capsule",
    )
    aux_3 = back_left_leg.body(name="aux_3", pos=[-0.2, -0.2, 0])
    aux_3.joint(
        axis=[0, 0, 1], name="hip_3", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_3.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="backleft_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_3 = aux_3.body(pos=[-0.2, -0.2, 0])
    ankle_3.joint(
        axis=[-1, 1, 0],
        name="ankle_3",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_3.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, -0.4, 0.0],
        name="backleft_ankle_geom",
        size="0.08",
        type="capsule",
    )

    back_right_leg = ant.body(name="back_right_leg", pos=[0, 0, 0])
    back_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="aux_4_geom",
        size="0.08",
        type="capsule",
    )
    aux_4 = back_right_leg.body(name="aux_4", pos=[0.2, -0.2, 0])
    aux_4.joint(
        axis=[0, 0, 1], name="hip_4", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_4.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="backright_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_4 = aux_4.body(pos=[0.2, -0.2, 0])
    ankle_4.joint(
        axis=[1, 1, 0],
        name="ankle_4",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_4.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, -0.4, 0.0],
        name="backright_ankle_geom",
        size="0.08",
        type="capsule",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_4", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_4", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_1", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_1", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_2", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_2", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_3", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_3", gear=gear)
    return mjcmodel


def angry_ant_crippled(gear=150):
    mjcmodel = MJCModel("ant_maze")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="degree", coordinate="local")
    mjcmodel.root.option(timestep="0.01", integrator="RK4")
    mjcmodel.root.custom().numeric(
        data="0.0 0.0 0.55 1.0 0.0 0.0 0.0 0.0 1.0 0.0 -1.0 0.0 -1.0 0.0 1.0",
        name="init_qpos",
    )
    asset = mjcmodel.root.asset()
    asset.texture(
        builtin="gradient",
        height="100",
        rgb1="1 1 1",
        rgb2="0 0 0",
        type="skybox",
        width="100",
    )
    asset.texture(
        builtin="flat",
        height="1278",
        mark="cross",
        markrgb="1 1 1",
        name="texgeom",
        random="0.01",
        rgb1="0.8 0.6 0.4",
        rgb2="0.8 0.6 0.4",
        type="cube",
        width="127",
    )
    asset.texture(
        builtin="checker",
        height="100",
        name="texplane",
        rgb1="0 0 0",
        rgb2="0.8 0.8 0.8",
        type="2d",
        width="100",
    )
    asset.material(
        name="MatPlane",
        reflectance="0.5",
        shininess="1",
        specular="1",
        texrepeat="60 60",
        texture="texplane",
    )
    asset.material(name="geom", texture="texgeom", texuniform="true")

    default = mjcmodel.root.default()
    default.joint(armature=1, damping=1, limited="true")
    default.geom(
        friction=[1.5, 0.5, 0.5],
        density=5.0,
        margin=0.01,
        condim=3,
        conaffinity=0,
        rgba="0.8 0.6 0.4 1",
    )

    worldbody = mjcmodel.root.worldbody()

    worldbody.geom(
        conaffinity=1,
        condim=3,
        material="MatPlane",
        name="floor",
        pos="0 0 0",
        rgba="0.8 0.9 0.8 1",
        size="40 40 40",
        type="plane",
    )
    worldbody.light(
        cutoff="100",
        diffuse=[0.8, 0.8, 0.8],
        dir="-0 0 -1.3",
        directional="true",
        exponent="1",
        pos="0 0 1.3",
        specular=".1 .1 .1",
    )

    ant = worldbody.body(name="torso", pos=[0, 0, 0.75])
    ant.geom(name="torso_geom", pos=[0, 0, 0], size="0.25", type="sphere")
    ant.joint(
        armature="0",
        damping="0",
        limited="false",
        margin="0.01",
        name="root",
        pos=[0, 0, 0],
        type="free",
    )

    eye_z = 0.1
    eye_y = -0.21
    eye_x_offset = 0.07
    # eyes
    ant.geom(
        fromto=[eye_x_offset, 0, eye_z, eye_x_offset, eye_y, eye_z],
        name="eye1",
        size="0.03",
        type="capsule",
        rgba=[1, 1, 1, 1],
    )
    ant.geom(
        fromto=[eye_x_offset, 0, eye_z, eye_x_offset, eye_y - 0.02, eye_z],
        name="eye1_",
        size="0.02",
        type="capsule",
        rgba=[0, 0, 0, 1],
    )
    ant.geom(
        fromto=[-eye_x_offset, 0, eye_z, -eye_x_offset, eye_y, eye_z],
        name="eye2",
        size="0.03",
        type="capsule",
        rgba=[1, 1, 1, 1],
    )
    ant.geom(
        fromto=[-eye_x_offset, 0, eye_z, -eye_x_offset, eye_y - 0.02, eye_z],
        name="eye2_",
        size="0.02",
        type="capsule",
        rgba=[0, 0, 0, 1],
    )
    # eyebrows
    ant.geom(
        fromto=[
            eye_x_offset - 0.03,
            eye_y,
            eye_z + 0.07,
            eye_x_offset + 0.03,
            eye_y,
            eye_z + 0.1,
        ],
        name="brow1",
        size="0.02",
        type="capsule",
        rgba=[0, 0, 0, 1],
    )
    ant.geom(
        fromto=[
            -eye_x_offset + 0.03,
            eye_y,
            eye_z + 0.07,
            -eye_x_offset - 0.03,
            eye_y,
            eye_z + 0.1,
        ],
        name="brow2",
        size="0.02",
        type="capsule",
        rgba=[0, 0, 0, 1],
    )

    front_left_leg = ant.body(name="front_left_leg", pos=[0, 0, 0])
    front_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="aux_1_geom",
        size="0.08",
        type="capsule",
    )
    aux_1 = front_left_leg.body(name="aux_1", pos=[0.2, 0.2, 0])
    aux_1.joint(
        axis=[0, 0, 1], name="hip_1", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="left_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_1 = aux_1.body(pos=[0.2, 0.2, 0])
    ankle_1.joint(
        axis=[-1, 1, 0],
        name="ankle_1",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, 0.4, 0.0],
        name="left_ankle_geom",
        size="0.08",
        type="capsule",
    )

    front_right_leg = ant.body(name="front_right_leg", pos=[0, 0, 0])
    front_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="aux_2_geom",
        size="0.08",
        type="capsule",
    )
    aux_2 = front_right_leg.body(name="aux_2", pos=[-0.2, 0.2, 0])
    aux_2.joint(
        axis=[0, 0, 1], name="hip_2", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="right_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_2 = aux_2.body(pos=[-0.2, 0.2, 0])
    ankle_2.joint(
        axis=[1, 1, 0],
        name="ankle_2",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, 0.4, 0.0],
        name="right_ankle_geom",
        size="0.08",
        type="capsule",
    )

    # Back left leg is crippled
    thigh_length = 0.1  # 0.2
    ankle_length = 0.2  # 0.4
    dark_red = [0.8, 0.3, 0.3, 1.0]

    back_left_leg = ant.body(name="back_left_leg", pos=[0, 0, 0])
    back_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="aux_3_geom",
        size="0.08",
        type="capsule",
        rgba=dark_red,
    )
    aux_3 = back_left_leg.body(name="aux_3", pos=[-0.2, -0.2, 0])
    aux_3.joint(
        axis=[0, 0, 1], name="hip_3", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_3.geom(
        fromto=[0.0, 0.0, 0.0, -thigh_length, -thigh_length, 0.0],
        name="backleft_leg_geom",
        size="0.08",
        type="capsule",
        rgba=dark_red,
    )
    ankle_3 = aux_3.body(pos=[-thigh_length, -thigh_length, 0])
    ankle_3.joint(
        axis=[-1, 1, 0],
        name="ankle_3",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_3.geom(
        fromto=[0.0, 0.0, 0.0, -ankle_length, -ankle_length, 0.0],
        name="backleft_ankle_geom",
        size="0.08",
        type="capsule",
        rgba=dark_red,
    )

    back_right_leg = ant.body(name="back_right_leg", pos=[0, 0, 0])
    back_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="aux_4_geom",
        size="0.08",
        type="capsule",
        rgba=dark_red,
    )
    aux_4 = back_right_leg.body(name="aux_4", pos=[0.2, -0.2, 0])
    aux_4.joint(
        axis=[0, 0, 1], name="hip_4", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_4.geom(
        fromto=[0.0, 0.0, 0.0, thigh_length, -thigh_length, 0.0],
        name="backright_leg_geom",
        size="0.08",
        type="capsule",
        rgba=dark_red,
    )
    ankle_4 = aux_4.body(pos=[thigh_length, -thigh_length, 0])
    ankle_4.joint(
        axis=[1, 1, 0],
        name="ankle_4",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_4.geom(
        fromto=[0.0, 0.0, 0.0, ankle_length, -ankle_length, 0.0],
        name="backright_ankle_geom",
        size="0.08",
        type="capsule",
        rgba=dark_red,
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_1", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_1", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_2", gear=gear)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_2", gear=gear)
    actuator.motor(
        ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_3", gear=1
    )  # cripple the joints
    actuator.motor(
        ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_3", gear=1
    )  # cripple the joints
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_4", gear=1)
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_4", gear=1)
    return mjcmodel


class CustomAntEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    """A modified ant env with lower joint gear ratios so it flips less often and learns
    faster."""

    def __init__(self, max_timesteps=1000, disabled=False, gear=150):
        self.timesteps = 0
        self.max_timesteps = max_timesteps

        if disabled:
            model = angry_ant_crippled(gear=gear)
        else:
            model = ant_env(gear=gear)

        with model.asfile() as f:
            mujoco_env.MujocoEnv.__init__(self, f.name, 5)

    def step(self, a):
        vel = self.sim.data.qvel.flat[0]
        forward_reward = vel
        self.do_simulation(a, self.frame_skip)

        ctrl_cost = 0.01 * np.square(a).sum()
        contact_cost = (
            0.5 * 1e-3 * np.sum(np.square(np.clip(self.sim.data.cfrc_ext, -1, 1)))
        )
        state = self.state_vector()
        flipped = not (state[2] >= 0.2)
        flipped_rew = -1 if flipped else 0
        reward = forward_reward - ctrl_cost - contact_cost + flipped_rew

        self.timesteps += 1
        done = self.timesteps >= self.max_timesteps

        ob = self._get_obs()
        return (
            ob,
            reward,
            done,
            dict(
                reward_forward=forward_reward,
                reward_ctrl=-ctrl_cost,
                reward_contact=-contact_cost,
                reward_flipped=flipped_rew,
            ),
        )

    def _get_obs(self):
        return np.concatenate(
            [
                self.sim.data.qpos.flat[2:],
                self.sim.data.qvel.flat,
                np.clip(self.sim.data.cfrc_ext, -1, 1).flat,
            ]
        )

    def reset_model(self):
        self.timesteps = 0
        qpos = self.init_qpos + self.np_random.uniform(
            size=self.model.nq, low=-0.1, high=0.1
        )
        qvel = self.init_qvel + self.np_random.randn(self.model.nv) * 0.1
        self.set_state(qpos, qvel)
        return self._get_obs()

    def viewer_setup(self):
        self.viewer.cam.distance = self.model.stat.extent * 0.5



"""Helper classes to serialize classes with TensorFlow objects."""

import os
import pickle
from abc import ABC, abstractmethod
from typing import Type, TypeVar

import tensorflow as tf

from imitation.util import networks

T = TypeVar("T", bound="Serializable")


class Serializable(ABC):
    """Abstract mix-in defining methods to load/save model."""

    @classmethod
    def load(cls: Type[T], directory: str) -> T:
        """Load object plus weights from directory."""
        with open(os.path.join(directory, "loader"), "rb") as f:
            load_cls = pickle.load(f)
        return load_cls._load(directory)

    def save(self, directory: str) -> None:
        """Save object and weights to directory."""
        os.makedirs(directory, exist_ok=True)

        load_path = os.path.join(directory, "loader")
        with open(load_path + ".tmp", "wb") as f:
            pickle.dump(type(self), f)
        # Ensure atomic write
        os.replace(load_path + ".tmp", load_path)

        self._save(directory)

    @classmethod
    @abstractmethod
    def _load(cls: Type[T], directory: str) -> T:
        """Class-specific loading logic."""

    @abstractmethod
    def _save(self, directory: str) -> None:
        """Class-specific saving logic."""


def make_cls(cls, args, kwargs):
    return cls(*args, **kwargs)


class LayersSerializable(Serializable):
    """Serialization mix-in based on `__init__` then rehydration.

    Subclasses must call the constructor with all arguments needed by `__init__`,
    and a dictionary mapping from strings to `tf.layers.Layer` objects.
    In most cases, you can use the following idiom::

        args = locals()
        # ... your normal constructor
        layers = # ... gather your TensorFlow objects
        LayersSerializable.__init__(**args, layers=layers)
    """

    def __init__(self, *args, layers: networks.LayersDict, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._checkpoint = tf.train.Checkpoint(**layers)

    def __reduce__(self):
        return (make_cls, (type(self), self._args, self._kwargs))

    def save_parameters(self, directory: str) -> None:
        file_prefix = os.path.join(directory, "weights")
        # TensorFlow stores a path to the latest checkpoint. It will use a relative
        # path if you give a relative `file_prefix`, otherwise it will use
        # absolute paths. We want it to use relative paths so saved models are
        # portable across machines. See TF issue #2973.
        file_prefix = os.path.relpath(file_prefix)
        self._checkpoint.save(file_prefix=file_prefix)

    def load_parameters(self, directory: str) -> None:
        restore = self._checkpoint.restore(tf.train.latest_checkpoint(directory))
        restore.assert_consumed().run_restore_ops()

    @classmethod
    def _load(cls: Type[T], directory: str) -> T:
        with open(os.path.join(directory, "obj"), "rb") as f:
            obj = pickle.load(f)
        assert isinstance(obj, cls)
        obj.load_parameters(directory)
        return obj

    def _save(self, directory: str) -> None:
        obj_path = os.path.join(directory, "obj")
        # obj is static, so no need to write them multiple times.
        # (Best to avoid it -- if we were die in the middle of save, it could
        # cause data corruption invalidating previous checkpoints.)
        if not os.path.exists(obj_path):
            # TODO(gleave): it'd be nice to support mutable Python state
            with open(obj_path, "wb") as f:
                pickle.dump(self, f)

        self.save_parameters(directory)



import os
from typing import Iterable, Tuple

from matplotlib import pyplot as plt


def save_fig(path: str, fig: plt.Figure, fmt: str = "pdf", dpi: int = 300, **kwargs):
    path = f"{path}.{fmt}"
    root_dir = os.path.dirname(path)
    os.makedirs(root_dir, exist_ok=True)
    kwargs.setdefault("transparent", True)
    with open(path, "wb") as f:
        fig.savefig(f, format=fmt, dpi=dpi, **kwargs)


def save_figs(
    root_dir: str, generator: Iterable[Tuple[str, plt.Figure]], **kwargs
) -> None:
    for name, fig in generator:
        name = name.replace("/", "_")
        path = os.path.join(root_dir, name)
        save_fig(path, fig, **kwargs)



"""Finite-horizon tabular Maximum Causal Entropy IRL.

Follows the description in chapters 9 and 10 of Brian Ziebart's `PhD thesis`_.

Uses NumPy-based optimizer Jax, so the code can be run without
PyTorch/TensorFlow, and some code for simple reward models.

.. _PhD thesis:
    http://www.cs.cmu.edu/~bziebart/publications/thesis-bziebart.pdf
"""

import abc

import jax
import jax.experimental.stax as jstax
import jax.numpy as jnp
import jax.random as jrandom
import numpy as np
import scipy
import tensorflow as tf


def mce_partition_fh(env, *, R=None):
    r"""Performs the soft Bellman backup for a finite-horizon, undiscounted MDP.

    Calculates V^{soft}, Q^{soft}, and pi using recurrences (9.1), (9.2), and
    (9.3) from Ziebart (2010).

    Args:
        env (ModelBasedEnv): a tabular, known-dynamics MDP.
        R (None or np.array): a reward matrix. Defaults to env.reward_matrix.

    Returns:
        (V, Q, \pi) corresponding to the soft values, Q-values and MCE policy.
        V is a 2d array, indexed V[t,s]. Q is a 3d array, indexed Q[t,s,a].
        \pi is a 3d array, indexed \pi[t,s,a].
    """

    # shorthand
    horizon = env.horizon
    n_states = env.n_states
    n_actions = env.n_actions
    T = env.transition_matrix
    if R is None:
        R = env.reward_matrix

    # Initialization
    # indexed as V[t,s]
    V = np.full((horizon, n_states), -np.inf)
    # indexed as Q[t,s,a]
    Q = np.zeros((horizon, n_states, n_actions))
    broad_R = R[:, None]

    # Base case: final timestep
    # final Q(s,a) is just reward
    Q[horizon - 1, :, :] = broad_R
    # V(s) is always normalising constant
    V[horizon - 1, :] = scipy.special.logsumexp(Q[horizon - 1, :, :], axis=1)

    # Recursive case
    for t in reversed(range(horizon - 1)):
        next_values_s_a = T @ V[t + 1, :]
        Q[t, :, :] = broad_R + next_values_s_a
        V[t, :] = scipy.special.logsumexp(Q[t, :, :], axis=1)

    pi = np.exp(Q - V[:, :, None])

    return V, Q, pi


def mce_occupancy_measures(env, *, R=None, pi=None):
    """Calculate state visitation frequency Ds for each state s under a given policy pi.

    You can get pi from `mce_partition_fh`.

    Args:
        env (ModelBasedEnv): a tabular MDP.
        R (None or np.ndarray): reward matrix. Defaults is env.reward_matrix.
        pi (None or np.ndarray): policy to simulate. Defaults to soft-optimal
            policy w.r.t reward matrix.

    Returns:
        Tuple of D (ndarray) and Dt (ndarray). D is an :math:`|S|`-dimensional
        vector recording the expected number of times each state is visited.
        Dt is a :math:`T*|S|`-dimensional vector recording the probability of
        being in a given state at a given timestep.
    """

    # shorthand
    horizon = env.horizon
    n_states = env.n_states
    n_actions = env.n_actions
    T = env.transition_matrix
    if R is None:
        R = env.reward_matrix
    if pi is None:
        _, _, pi = mce_partition_fh(env, R=R)

    D = np.zeros((horizon, n_states))
    D[0, :] = env.initial_state_dist
    for t in range(1, horizon):
        for a in range(n_actions):
            E = D[t - 1] * pi[t - 1, :, a]
            D[t, :] += E @ T[:, a, :]

    return D, D.sum(axis=0)


def mce_irl(
    env,
    optimiser_tuple,
    rmodel,
    demo_state_om,
    linf_eps=1e-3,
    grad_l2_eps=1e-4,
    print_interval=100,
):
    r"""Tabular MCE IRL.

    Args:
        env (ModelBasedEnv): a tabular MDP.
        optimiser_tuple (tuple): a tuple of `(optim_init_fn, optim_update_fn,
            get_params_fn)` produced by a Jax optimiser.
        rmodel (RewardModel): a reward function to be optimised.
        demo_state_om (np.ndarray): matrix representing state occupancy measure
            for demonstrator.
        linf_eps (float): optimisation terminates if the $l_{\infty}$ distance
            between the demonstrator's state occupancy measure and the state
            occupancy measure for the current reward falls below this value.
        grad_l2_eps (float): optimisation also terminates if the $\ell_2$ norm
            of the MCE IRL gradient falls below this value.
        print_interval (int or None): how often to log current loss stats
            (using tf.logging). None to disable.

    Returns:
        (np.ndarray, np.ndarray): tuple of final parameters found by optimiser
        and state occupancy measure for the final reward function. Note that
        rmodel will also be updated with the latest parameters."""

    obs_mat = env.observation_matrix
    # l_\infty distance between demonstrator occupancy measure (OM) and OM for
    # soft-optimal policy w.r.t current reward (initially set to this value to
    # prevent termination)
    linf_delta = linf_eps + 1
    # norm of the MCE IRL gradient (also set to this value to prevent
    # termination)
    grad_norm = grad_l2_eps + 1
    # number of optimisation steps taken
    t = 0
    assert demo_state_om.shape == (len(obs_mat),)
    opt_init, opt_update, opt_get_params = optimiser_tuple
    rew_params = rmodel.get_params()
    opt_state = opt_init(rew_params)

    while linf_delta > linf_eps and grad_norm > grad_l2_eps:
        # get reward predicted for each state by current model, & compute
        # expected # of times each state is visited by soft-optimal policy
        # w.r.t that reward function
        predicted_r, out_grads = rmodel.out_grads(obs_mat)
        _, visitations = mce_occupancy_measures(env, R=predicted_r)
        # gradient of partition function w.r.t parameters; equiv to expectation
        # over states drawn from current imitation distribution of the gradient
        # of the reward function w.r.t its params
        pol_grad = np.sum(visitations[:, None] * out_grads, axis=0)
        # gradient of reward function w.r.t parameters, with expectation taken
        # over states
        expert_grad = np.sum(demo_state_om[:, None] * out_grads, axis=0)
        grad = pol_grad - expert_grad

        # these are just for termination conditions & debug logging
        grad_norm = np.linalg.norm(grad)
        linf_delta = np.max(np.abs(demo_state_om - visitations))
        if print_interval is not None and 0 == (t % print_interval):
            tf.logging.info(
                "Occupancy measure error@iter % 3d: %f (||params||=%f, "
                "||grad||=%f, ||E[dr/dw]||=%f)"
                % (
                    t,
                    linf_delta,
                    np.linalg.norm(rew_params),
                    np.linalg.norm(grad),
                    np.linalg.norm(pol_grad),
                )
            )

        # take a single optimiser step
        opt_state = opt_update(t, grad, opt_state)
        rew_params = opt_get_params(opt_state)
        rmodel.set_params(rew_params)
        t += 1

    return rew_params, visitations


# ############################### #
# ####### REWARD MODELS ######### #
# ############################### #


class RewardModel(abc.ABC):
    """Abstract model for reward functions (linear, MLPs, nearest-neighbour, etc.)"""

    @abc.abstractmethod
    def out(self, inputs):
        """Get rewards for a batch of observations.

        Args:
            inputs (np.ndarray): 2D matrix of observations, with first axis
                most likely indexing over state & second indexing over elements
                of observations themselves.

        Returns:
            np.ndarray of rewards (just a 1D vector with one element for each
            supplied observation).
        """

    @abc.abstractmethod
    def grads(self, inputs):
        """Gradients of reward with respect to a batch of input observations.

        Args:
            inputs (np.ndarray): 2D matrix of observations, like .out().

        Returns:
            np.ndarray of gradients *with respect to each input separately*.
            e.g if the model has a W-dimensional parameter vector, and there
            are O observation passed in, then the return value will be an O*W
            matrix of gradients.
        """

    def out_grads(self, inputs):
        """Combination method to do forward-prop AND back-prop. This is trivial for
        linear models, but might provide some cost saving for deep ones.

        Args:
            inputs (np.ndarray): 2D matrix of observations, like .out().

        Returns:
            (np.ndarray, np.ndarray), where first array is equivalent to return
            value of .out() and second array is equivalent to return value of
            .grads().
        """
        return self.out(inputs), self.grads(inputs)

    @abc.abstractmethod
    def set_params(self, params):
        """Set a new parameter vector for the model (from flat Numpy array).

        Args:
            params (np.ndarray): 1D parameter vector for the model.
        """

    @abc.abstractmethod
    def get_params(self):
        """Get current parameter vector from model (as flat Numpy array).

        Args: empty.

        Returns:
            np.ndarray: 1D parameter vector for the model.
        """


class LinearRewardModel(RewardModel):
    """Linear reward model (without bias)."""

    def __init__(self, obs_dim, *, seed=None):
        """Construct linear reward model for `obs_dim`-dimensional observation space.

        Initial values are generated from given seed (int or None).

        Args:
            obs_dim (int): dimensionality of observation space.
            seed (int or None): random seed for generating initial params. If
                None, seed will be chosen arbitrarily
        """
        if seed is not None:
            rng = np.random.RandomState(seed)
        else:
            rng = np.random
        self._weights = rng.randn(obs_dim)

    def out(self, inputs):
        assert inputs.shape[1:] == self._weights.shape
        return inputs @ self._weights

    def grads(self, inputs):
        assert inputs.shape[1:] == self._weights.shape
        return inputs

    def set_params(self, params):
        assert params.shape == self._weights.shape
        self._weights = params

    def get_params(self):
        return self._weights


class JaxRewardModel(RewardModel, abc.ABC):
    """Wrapper for arbitrary Jax-based reward models.

    Useful for neural nets.
    """

    def __init__(self, obs_dim, *, seed=None):
        """Internal setup for Jax-based reward models.

        Initialises reward model using given seed & input size (`obs_dim`).

        Args:
            obs_dim (int): dimensionality of observation space.
            seed (int or None): random seed for generating initial params. If
                None, seed will be chosen arbitrarily, as in
                LinearRewardModel.
        """
        # TODO: apply jax.jit() to everything in sight
        net_init, self._net_apply = self.make_stax_model()
        if seed is None:
            # oh well
            seed = np.random.randint((1 << 63) - 1)
        rng = jrandom.PRNGKey(seed)
        out_shape, self._net_params = net_init(rng, (-1, obs_dim))
        self._net_grads = jax.grad(self._net_apply)
        # output shape should just be batch dim, nothing else
        assert out_shape == (-1,), "got a weird output shape %s" % (out_shape,)

    @abc.abstractmethod
    def make_stax_model(self):
        """Build the stax model that this thing is meant to optimise.

        Should return (net_init, net_apply) pair, just like Stax modules.

        Returns:
            tuple of net_init(rng, input_shape) function to initialise the
            network, and net_apply(params, inputs) to do forward prop on the
            network.
        """

    def _flatten(self, matrix_tups):
        """Flatten everything and concatenate it together."""
        out_vecs = [v.flatten() for t in matrix_tups for v in t]
        return jnp.concatenate(out_vecs)

    def _flatten_batch(self, matrix_tups):
        """Flatten all except leading dim & concatenate results together in channel dim.

        (Channel dim is whatever the dim after the leading dim is)."""
        out_vecs = []
        for t in matrix_tups:
            for v in t:
                new_shape = (v.shape[0],)
                if len(v.shape) > 1:
                    new_shape = new_shape + (np.prod(v.shape[1:]),)
                out_vecs.append(v.reshape(new_shape))
        return jnp.concatenate(out_vecs, axis=1)

    def out(self, inputs):
        return np.asarray(self._net_apply(self._net_params, inputs))

    def grads(self, inputs):
        in_grad_partial = jax.partial(self._net_grads, self._net_params)
        grad_vmap = jax.vmap(in_grad_partial)
        rich_grads = grad_vmap(inputs)
        flat_grads = np.asarray(self._flatten_batch(rich_grads))
        assert flat_grads.ndim == 2 and flat_grads.shape[0] == inputs.shape[0]
        return flat_grads

    def set_params(self, params):
        # have to reconstitute appropriately-shaped weights from 1D param vec
        # shit this is going to be annoying
        idx_acc = 0
        new_params = []
        for t in self._net_params:
            new_t = []
            for v in t:
                new_idx_acc = idx_acc + v.size
                new_v = params[idx_acc:new_idx_acc].reshape(v.shape)
                # this seems to cast it to Jax DeviceArray appropriately;
                # surely there's better way, though?
                new_v = 0.0 * v + new_v
                new_t.append(new_v)
                idx_acc = new_idx_acc
            new_params.append(new_t)
        self._net_params = new_params

    def get_params(self):
        return self._flatten(self._net_params)


class MLPRewardModel(JaxRewardModel):
    """Simple MLP-based reward function with Jax/Stax."""

    def __init__(self, obs_dim, hiddens, activation="Tanh", **kwargs):
        """Construct an MLP-based reward function.

        Args:
            obs_dim (int): dimensionality of observation space.
            hiddens ([int]): size of hidden layers.
            activation (str): name of activation (Tanh, Relu, Softplus
                supported).
            **kwargs: extra keyword arguments to be passed to
                JaxRewardModel.__init__().
        """
        assert activation in ["Tanh", "Relu", "Softplus"], (
            "probably can't handle activation '%s'" % activation
        )
        self._hiddens = hiddens
        self._activation = activation
        super().__init__(obs_dim, **kwargs)

    def make_stax_model(self):
        act = getattr(jstax, self._activation)
        layers = []
        for h in self._hiddens:
            layers.extend([jstax.Dense(h), act])
        layers.extend([jstax.Dense(1), _StaxSqueeze()])
        return jstax.serial(*layers)


def _StaxSqueeze(axis=-1):
    """Stax layer that collapses a single axis that has dimension 1.

    Only used in MLPRewardModel.
    """

    def init_fun(rng, input_shape):
        ax = axis
        if ax < 0:
            ax = len(input_shape) + ax
        assert ax < len(input_shape), "invalid axis %d for %d-dimensional tensor" % (
            axis,
            len(input_shape),
        )
        assert input_shape[ax] == 1, "axis %d is %d, not 1" % (axis, input_shape[ax])
        output_shape = input_shape[:ax] + input_shape[ax + 1 :]
        return output_shape, ()

    def apply_fun(params, inputs, **kwargs):
        return jnp.squeeze(inputs, axis=axis)

    return init_fun, apply_fun



"""Config files for parallel experiments.

Parallel experiments are intended to be defined in Python rather than
via CLI. For example, a user should add a new
`@parallel_ex.named_config` to define a new parallel experiment.

Adding custom named configs is necessary because the CLI interface can't add
search spaces to the config like `"seed": tune.grid_search([0, 1, 2, 3])`.
"""

import numpy as np
import ray.tune as tune
import sacred

from imitation.util.util import make_unique_timestamp

parallel_ex = sacred.Experiment("parallel")


@parallel_ex.config
def config():
    sacred_ex_name = "expert_demos"  # The experiment to parallelize
    init_kwargs = {}  # Keyword arguments to pass to ray.init()
    _uuid = make_unique_timestamp()
    run_name = f"DEFAULT_{_uuid}"  # CLI --name option. For analysis grouping.
    resources_per_trial = {}  # Argument to `tune.run`
    base_named_configs = []  # Background settings before search_space is applied
    base_config_updates = {}  # Background settings before search_space is applied
    search_space = {
        "named_configs": [],
        "config_updates": {},
    }  # `config` argument to `ray.tune.run(trainable, config)`

    local_dir = None  # `local_dir` arg for `ray.tune.run`
    upload_dir = None  # `upload_dir` arg for `ray.tune.run`
    n_seeds = 3  # Number of seeds to search over by default


@parallel_ex.config
def seeds(n_seeds):
    search_space = {"config_updates": {"seed": tune.grid_search(list(range(n_seeds)))}}


@parallel_ex.named_config
def s3():
    upload_dir = "s3://shwang-chai/private"


# Debug named configs
@parallel_ex.named_config
def debug_log_root():
    search_space = {"config_updates": {"log_root": "/tmp/debug_parallel_ex"}}


@parallel_ex.named_config
def generate_test_data():
    """Used by tests/generate_test_data.sh to generate tests/data/gather_tb/.

    "tests/data/gather_tb/" should contain 4 Tensorboard run directories ("sb_tb/" and
    "tb/" for each of two trials in the search space below).
    """
    sacred_ex_name = "expert_demos"
    run_name = "TEST"
    n_seeds = 1
    search_space = {
        "config_updates": {
            "init_rl_kwargs": {
                "learning_rate": tune.grid_search([3e-4 * x for x in (1 / 3, 1 / 2)]),
            },
        }
    }
    base_named_configs = ["cartpole", "fast"]
    base_config_updates = {
        "init_tensorboard": True,
        "rollout_save_final": False,
    }


@parallel_ex.named_config
def example_cartpole_rl():
    sacred_ex_name = "expert_demos"
    run_name = "example-cartpole"
    n_seeds = 2
    search_space = {
        "config_updates": {
            "init_rl_kwargs": {
                "learning_rate": tune.grid_search(np.logspace(3e-6, 1e-1, num=3)),
                "nminibatches": tune.grid_search([16, 32, 64]),
            },
        }
    }
    base_named_configs = ["cartpole"]
    base_config_updates = {"init_tensorboard": True}
    resources_per_trial = dict(cpu=4)


EASY_ENVS = ["cartpole", "pendulum", "mountain_car"]


@parallel_ex.named_config
def example_rl_easy():
    sacred_ex_name = "expert_demos"
    run_name = "example-rl-easy"
    n_seeds = 2
    search_space = {
        "named_configs": tune.grid_search([[env] for env in EASY_ENVS]),
        "config_updates": {
            "init_rl_kwargs": {
                "learning_rate": tune.grid_search(np.logspace(3e-6, 1e-1, num=3)),
                "nminibatches": tune.grid_search([16, 32, 64]),
            },
        },
    }
    base_config_updates = {"init_tensorboard": True}
    resources_per_trial = dict(cpu=4)


@parallel_ex.named_config
def example_gail_easy():
    sacred_ex_name = "train_adversarial"
    run_name = "example-gail-easy"
    n_seeds = 1
    search_space = {
        "named_configs": tune.grid_search([[env] for env in EASY_ENVS]),
        "config_updates": {
            "init_trainer_kwargs": {
                "init_rl_kwargs": {
                    "learning_rate": tune.grid_search(np.logspace(3e-6, 1e-1, num=3)),
                    "nminibatches": tune.grid_search([16, 32, 64]),
                },
            },
        },
    }
    base_config_updates = {
        "init_tensorboard": True,
        "init_trainer_kwargs": {"use_gail": True},
    }



"""Example discrete MDPs for use with tabular MCE IRL."""

import gym
import numpy as np

from imitation.envs.resettable_env import TabularModelEnv


def make_random_trans_mat(n_states, n_actions, max_branch_factor, rand_state=np.random):
    """Make a 'random' transition matrix.

    Each action goes to at least `max_branch_factor` other states from the
    current state, with transition distribution sampled from Dirichlet(1,1,…,1).

    This roughly apes the strategy from some old Lisp code that Rich Sutton
    left on the internet (http://incompleteideas.net/RandomMDPs.html), and is
    therefore a legitimate way to generate MDPs.
    """
    out_mat = np.zeros((n_states, n_actions, n_states), dtype="float32")
    for start_state in range(n_states):
        for action in range(n_actions):
            # uniformly sample a number of successors in [1,max_branch_factor]
            # for this action
            succs = rand_state.randint(1, max_branch_factor + 1)
            next_states = rand_state.choice(n_states, size=(succs,), replace=False)
            # generate random vec in probability simplex
            next_vec = rand_state.dirichlet(np.ones((succs,)))
            next_vec = next_vec / np.sum(next_vec)
            out_mat[start_state, action, next_states] = next_vec
    return out_mat


def make_random_state_dist(n_avail, n_states, rand_state=np.random):
    """Make a random initial state distribution over n_states in which n_avail<=n_states
    of the states are supported."""
    assert 0 < n_avail <= n_states
    init_dist = np.zeros((n_states,))
    next_states = rand_state.choice(n_states, size=(n_avail,), replace=False)
    avail_state_dist = rand_state.dirichlet(np.ones((n_avail,)))
    init_dist[next_states] = avail_state_dist
    assert np.sum(init_dist > 0) == n_avail
    init_dist = init_dist / np.sum(init_dist)
    return init_dist


def make_obs_mat(n_states, is_random, obs_dim, rand_state=np.random):
    """Makes an observation matrix with a single observation for each state.

    Args:
        n_states (int): Number of states.
        is_random (bool): Are observations drawn at random?
                    If `True`, draw from random normal distribution.
                    If `False`, are unique one-hot vectors for each state.
        obs_dim (int or NoneType): Must be `None` if `is_random == False`.
                 Otherwise, this must be set to the size of the random vectors.
        rand_state (np.random.RandomState): Random number generator.

    Returns:
        A matrix of shape `(n_states, obs_dim if is_random else n_states)`.
    """
    if not is_random:
        assert obs_dim is None
    if is_random:
        obs_mat = rand_state.normal(0, 2, (n_states, obs_dim))
    else:
        obs_mat = np.identity(n_states)
    assert (
        obs_mat.ndim == 2 and obs_mat.shape[:1] == (n_states,) and obs_mat.shape[1] > 0
    )
    return obs_mat


class RandomMDP(TabularModelEnv):
    """AN MDP with a random transition matrix.

    Random matrix is created by `make_random_trans_mat`.
    """

    def __init__(
        self,
        n_states,
        n_actions,
        branch_factor,
        horizon,
        random_obs,
        *,
        obs_dim=None,
        generator_seed=None,
    ):
        super().__init__()
        # this generator is ONLY for constructing the MDP, not for controlling
        # random outcomes during rollouts
        rand_gen = np.random.RandomState(generator_seed)
        if random_obs:
            if obs_dim is None:
                obs_dim = n_states
        else:
            assert obs_dim is None
        self._observation_matrix = make_obs_mat(
            n_states=n_states,
            is_random=random_obs,
            obs_dim=obs_dim,
            rand_state=rand_gen,
        )
        self._transition_matrix = make_random_trans_mat(
            n_states=n_states,
            n_actions=n_actions,
            max_branch_factor=branch_factor,
            rand_state=rand_gen,
        )
        self._initial_state_dist = make_random_state_dist(
            n_avail=branch_factor, n_states=n_states, rand_state=rand_gen
        )
        self._horizon = horizon
        self._reward_weights = rand_gen.randn(self._observation_matrix.shape[-1])
        self._reward_matrix = self._observation_matrix @ self._reward_weights
        assert self._reward_matrix.shape == (self.n_states,)

    @property
    def observation_matrix(self):
        return self._observation_matrix

    @property
    def transition_matrix(self):
        return self._transition_matrix

    @property
    def reward_matrix(self):
        return self._reward_matrix

    @property
    def initial_state_dist(self):
        return self._initial_state_dist

    @property
    def horizon(self):
        return self._horizon


class CliffWorld(TabularModelEnv):
    """A grid world like this::

         0 1 2 3 4 5 6 7 8 9
        +-+-+-+-+-+-+-+-+-+-+  Wind:
      0 |S|C|C|C|C|C|C|C|C|G|
        +-+-+-+-+-+-+-+-+-+-+  ^ ^ ^
      1 | | | | | | | | | | |  | | |
        +-+-+-+-+-+-+-+-+-+-+
      2 | | | | | | | | | | |  ^ ^ ^
        +-+-+-+-+-+-+-+-+-+-+  | | |

    Aim is to get from S to G. The G square has reward +10, the C squares
    ("cliff") have reward -10, and all other squares have reward -1. Agent can
    move in all directions (except through walls), but there is 30% chance that
    they will be blown upwards by one more unit than intended due to wind.
    Optimal policy is to go out a bit and avoid the cliff, but still hit goal
    eventually.
    """

    def __init__(
        self,
        width,
        height,
        horizon,
        use_xy_obs,
        *,
        rew_default=-1,
        rew_goal=10,
        rew_cliff=-10,
        fail_p=0.3,
    ):
        super().__init__()
        assert (
            width >= 3 and height >= 2
        ), "degenerate grid world requested; is this a bug?"
        self.width = width
        self.height = height
        succ_p = 1 - fail_p
        n_states = width * height
        O_mat = self._observation_matrix = np.zeros(
            (n_states, 2 if use_xy_obs else n_states)
        )
        R_vec = self._reward_matrix = np.zeros((n_states,))
        T_mat = self._transition_matrix = np.zeros((n_states, 4, n_states))
        self._horizon = horizon

        def to_id_clamp(row, col):
            """Convert (x,y) state to state ID, after clamp x & y to lie in grid."""
            row = min(max(row, 0), height - 1)
            col = min(max(col, 0), width - 1)
            state_id = row * width + col
            assert 0 <= state_id < self.n_states
            return state_id

        for row in range(height):
            for col in range(width):
                state_id = to_id_clamp(row, col)

                # start by computing reward
                if row > 0:
                    r = rew_default  # blank
                elif col == 0:
                    r = rew_default  # start
                elif col == width - 1:
                    r = rew_goal  # goal
                else:
                    r = rew_cliff  # cliff
                R_vec[state_id] = r

                # now compute observation
                if use_xy_obs:
                    # (x, y) coordinate scaled to (0,1)
                    O_mat[state_id, :] = [
                        float(col) / (width - 1),
                        float(row) / (height - 1),
                    ]
                else:
                    # our observation matrix is just the identity; observation
                    # is an indicator vector telling us exactly what state
                    # we're in
                    O_mat[state_id, state_id] = 1

                # finally, compute transition matrix entries for each of the
                # four actions
                for drow in [-1, 1]:
                    for dcol in [-1, 1]:
                        action_id = (drow + 1) + (dcol + 1) // 2
                        target_state = to_id_clamp(row + drow, col + dcol)
                        fail_state = to_id_clamp(row + drow - 1, col + dcol)
                        T_mat[state_id, action_id, fail_state] += fail_p
                        T_mat[state_id, action_id, target_state] += succ_p

        assert np.allclose(np.sum(T_mat, axis=-1), 1, rtol=1e-5), (
            "un-normalised matrix %s" % O_mat
        )

    @property
    def observation_matrix(self):
        return self._observation_matrix

    @property
    def transition_matrix(self):
        return self._transition_matrix

    @property
    def reward_matrix(self):
        return self._reward_matrix

    @property
    def horizon(self):
        return self._horizon

    @property
    def initial_state_dist(self):
        # always start in s0
        rv = np.zeros((self.n_states,))
        rv[0] = 1.0
        return rv

    def draw_value_vec(self, D):
        """Use matplotlib a vector of values for each state.

        The vector could represent things like reward, occupancy measure, etc.
        """
        import matplotlib.pyplot as plt

        grid = D.reshape(self.height, self.width)
        plt.imshow(grid)
        plt.gca().grid(False)


def register_cliff(suffix, kwargs):
    gym.register(
        f"imitation/CliffWorld{suffix}-v0",
        entry_point="imitation.envs.examples.model_envs:CliffWorld",
        kwargs=kwargs,
    )


for width, height, horizon in [(7, 4, 9), (15, 6, 18), (100, 20, 110)]:
    for use_xy in [False, True]:
        use_xy_str = "XY" if use_xy else ""
        register_cliff(
            f"{width}x{height}{use_xy_str}",
            kwargs={
                "width": width,
                "height": height,
                "use_xy_obs": use_xy,
                "horizon": horizon,
            },
        )

# These parameter choices are somewhat arbitrary.
# We anticipate most users will want to construct RandomMDP directly.
gym.register(
    "imitation/Random-v0",
    entry_point="imitation.envs.examples.model_envs:RandomMDP",
    kwargs={
        "n_states": 16,
        "n_actions": 3,
        "branch_factor": 2,
        "horizon": 20,
        "random_obs": True,
        "obs_dim": 5,
        "generator_seed": 42,
    },
)



from typing import List

import gym
import numpy as np
from stable_baselines.common.vec_env import VecEnv, VecEnvWrapper

from imitation.data import rollout, types


class BufferingWrapper(VecEnvWrapper):
    """Saves transitions of underlying VecEnv.

    Retrieve saved transitions using `pop_transitions()`.
    """

    def __init__(self, venv: VecEnv, error_on_premature_reset: bool = True):
        """
        Args:
          venv: The wrapped VecEnv.
          error_on_premature_reset: Error if `reset()` is called on this wrapper
            and there are saved samples that haven't yet been accessed.
        """
        super().__init__(venv)
        self.error_on_premature_reset = error_on_premature_reset
        self._trajectories = []
        self._init_reset = False
        self._traj_accum = None
        self._saved_acts = None
        self.n_transitions = None

    def reset(self, **kwargs):
        if (
            self._init_reset
            and self.error_on_premature_reset
            and self.n_transitions > 0
        ):  # noqa: E127
            raise RuntimeError("BufferingWrapper reset() before samples were accessed")
        self._init_reset = True
        self.n_transitions = 0
        obs = self.venv.reset(**kwargs)
        self._traj_accum = rollout.TrajectoryAccumulator()
        for i, ob in enumerate(obs):
            self._traj_accum.add_step({"obs": ob}, key=i)
        return obs

    def step_async(self, actions):
        assert self._init_reset
        assert self._saved_acts is None
        self.venv.step_async(actions)
        self._saved_acts = actions

    def step_wait(self):
        assert self._init_reset
        assert self._saved_acts is not None
        acts, self._saved_acts = self._saved_acts, None
        obs, rews, dones, infos = self.venv.step_wait()
        finished_trajs = self._traj_accum.add_steps_and_auto_finish(
            acts, obs, rews, dones, infos
        )
        self._trajectories.extend(finished_trajs)
        self.n_transitions += self.num_envs
        return obs, rews, dones, infos

    def _finish_partial_trajectories(self) -> List[types.TrajectoryWithRew]:
        """Finishes and returns partial trajectories in `self._traj_accum`."""
        trajs = []
        for i in range(self.num_envs):
            # Check that we have any transitions at all.
            # The number of "transitions" or "timesteps" stored for the ith
            # environment is the number of step dicts stored in
            # `partial_trajectories[i]` minus one. We need to offset by one because
            # the first step dict is comes from `reset()`, not from `step()`.
            n_transitions = len(self._traj_accum.partial_trajectories[i]) - 1
            assert n_transitions >= 0, "Invalid TrajectoryAccumulator state"
            if n_transitions >= 1:
                traj = self._traj_accum.finish_trajectory(i)
                trajs.append(traj)

                # Reinitialize a partial trajectory starting with the final observation.
                self._traj_accum.add_step({"obs": traj.obs[-1]}, key=i)
        return trajs

    def pop_transitions(self) -> types.TransitionsWithRew:
        """Pops recorded transitions, returning them as an instance of Transitions.

        Raises a RuntimeError if called when `self.n_transitions == 0`.
        """
        if self.n_transitions == 0:
            # It would be better to return an empty `Transitions`, but we would need
            # to get the non-zero dimensions of every np.ndarray attribute correct to
            # avoid downstream errors. This is easier and sufficient for now.
            raise RuntimeError("Called pop_transitions on an empty BufferingWrapper")
        partial_trajs = self._finish_partial_trajectories()
        self._trajectories.extend(partial_trajs)
        transitions = rollout.flatten_trajectories_with_rew(self._trajectories)
        assert len(transitions.obs) == self.n_transitions
        self._trajectories = []
        self.n_transitions = 0
        return transitions


class RolloutInfoWrapper(gym.Wrapper):
    """Add the entire episode's rewards and observations to `info` at episode end.

    Whenever done=True, `info["rollouts"]` is a dict with keys "obs" and "rews", whose
    corresponding values hold the Numpy arrays containing the raw observations and
    rewards seen during this episode.
    """

    def __init__(self, env):
        super().__init__(env)
        self._obs = None
        self._rews = None

    def reset(self, **kwargs):
        new_obs = super().reset()
        self._obs = [new_obs]
        self._rews = []
        return new_obs

    def step(self, action):
        obs, rew, done, info = self.env.step(action)
        self._obs.append(obs)
        self._rews.append(rew)

        if done:
            assert "rollout" not in info
            info["rollout"] = {
                "obs": np.stack(self._obs),
                "rews": np.stack(self._rews),
            }
        return obs, rew, done, info



# pytype: skip-file
# flake8: noqa

import warnings
from typing import Dict, Generic, List, Optional, TypeVar

import gym
import numpy as np

from imitation.data import buffer, datasets, types

S = TypeVar("S")
T = TypeVar("T")


"""Test: Add scaling wrapper that just multiplies obs and acts by K, M."""


class TransitionsDatasetWrapper(datasets.Dataset[types.TransitionsWithRew]):
    """Wrapper that converts dictionary samples to TransitionsWithRew.

    The dictionary samples must have the same keys as TransitionsWithRew's fields.
    Namely, 'obs', 'acts', 'next_obs', 'rews', and 'dones'.

    Useful for GAIL and AIRL, which require `Dataset[imitation.data.types.Transitions]`
    instead of `Dataset[BatchDict]`.
    """

    def __init__(self, minerl_dataset: datasets.Dataset[Dict[str, np.ndarray]]):
        f"""Constructs DictToTransitionsWrapper."""
        self.dataset = minerl_dataset

    def sample(self, n_samples: int) -> types.TransitionsWithRew:
        return types.TransitionsWithRew(**self.dataset.sample(n_samples))

    def size(self) -> Optional[int]:
        return self.dataset.size()



import numpy as np


def flat_to_one_hot(val, ndim):
    """

    >>> flat_to_one_hot(2, ndim=4)
    array([ 0.,  0.,  1.,  0.])
    >>> flat_to_one_hot(4, ndim=5)
    array([ 0.,  0.,  0.,  0.,  1.])
    >>> flat_to_one_hot(np.array([2, 4, 3]), ndim=5)
    array([[ 0.,  0.,  1.,  0.,  0.],
           [ 0.,  0.,  0.,  0.,  1.],
           [ 0.,  0.,  0.,  1.,  0.]])
    """
    shape = np.array(val).shape
    v = np.zeros(shape + (ndim,))
    if len(shape) == 1:
        v[np.arange(shape[0]), val] = 1.0
    else:
        v[val] = 1.0
    return v


def one_hot_to_flat(val):
    """
    >>> one_hot_to_flat(np.array([0,0,0,0,1]))
    4
    >>> one_hot_to_flat(np.array([0,0,1,0]))
    2
    >>> one_hot_to_flat(np.array([[0,0,1,0], [1,0,0,0], [0,1,0,0]]))
    array([2, 0, 1])
    """
    idxs = np.array(np.where(val == 1.0))[-1]
    if len(val.shape) == 1:
        return int(idxs)
    return idxs



"""Load serialized policies of different types."""

import contextlib
import os
import pickle
from typing import Callable, ContextManager, Iterator, Optional, Type

import tensorflow as tf
from stable_baselines.common.base_class import BaseRLModel
from stable_baselines.common.policies import BasePolicy
from stable_baselines.common.vec_env import VecEnv, VecNormalize

from imitation.policies.base import RandomPolicy, ZeroPolicy
from imitation.util import registry

PolicyLoaderFn = Callable[[str, VecEnv], ContextManager[BasePolicy]]

policy_registry: registry.Registry[PolicyLoaderFn] = registry.Registry()


class NormalizePolicy(BasePolicy):
    """Wraps a policy, normalizing its input observations.

    `VecNormalize` normalizes observations to have zero mean and unit standard
    deviation. To do this, it collects statistics on the observations. We must
    restore these statistics when we load the policy, or we will be feeding
    observations in of a different scale to those the policy was trained with.

    It is convenient to do this when loading the policy, so users of a saved
    policy are not responsible for this implementation detail. WARNING: This
    trick will not work for fine-tuning / training policies.
    """

    def __init__(self, policy: BasePolicy, vec_normalize: VecNormalize):
        super().__init__(
            policy.sess,
            policy.ob_space,
            policy.ac_space,
            policy.n_env,
            policy.n_steps,
            policy.n_batch,
        )
        self._policy = policy
        self.vec_normalize = vec_normalize

    def _wrapper(self, fn, obs, state=None, mask=None, *args, **kwargs):
        norm_obs = self.vec_normalize.normalize_obs(obs)
        return fn(norm_obs, state=state, mask=mask, *args, **kwargs)

    def step(self, *args, **kwargs):
        return self._wrapper(self._policy.step, *args, **kwargs)

    def proba_step(self, *args, **kwargs):
        return self._wrapper(self._policy.proba_step, *args, **kwargs)


def _load_stable_baselines(cls: Type[BaseRLModel], policy_attr: str) -> PolicyLoaderFn:
    """Higher-order function, returning a policy loading function.

    Args:
        cls: The RL algorithm, e.g. `stable_baselines.PPO2`.
        policy_attr: The attribute of the RL algorithm containing the policy,
            e.g. `act_model`.

    Returns:
        A function loading policies trained via cls.
    """

    @contextlib.contextmanager
    def f(path: str, venv: VecEnv) -> Iterator[BasePolicy]:
        """Loads a policy saved to path, for environment env."""
        tf.logging.info(
            f"Loading Stable Baselines policy for '{cls}' " f"from '{path}'"
        )
        model_path = os.path.join(path, "model.pkl")
        model = None
        try:
            model = cls.load(model_path, env=venv)
            policy = getattr(model, policy_attr)

            try:
                normalize_path = os.path.join(path, "vec_normalize.pkl")
                with open(normalize_path, "rb") as f:
                    vec_normalize = pickle.load(f)
                vec_normalize.training = False
                vec_normalize.set_venv(venv)
                policy = NormalizePolicy(policy, vec_normalize)
                tf.logging.info(f"Loaded VecNormalize from '{normalize_path}'")
            except FileNotFoundError:
                # We did not use VecNormalize during training, skip
                pass

            yield policy
        finally:
            if model is not None and model.sess is not None:
                model.sess.close()

    return f


policy_registry.register(
    "random",
    value=registry.build_loader_fn_require_space(
        registry.dummy_context(RandomPolicy),
    ),
)
policy_registry.register(
    "zero",
    value=registry.build_loader_fn_require_space(
        registry.dummy_context(ZeroPolicy),
    ),
)


def _add_stable_baselines_policies(classes):
    for k, (cls_name, attr) in classes.items():
        try:
            cls = registry.load_attr(cls_name)
            fn = _load_stable_baselines(cls, attr)
            policy_registry.register(k, value=fn)
        except (AttributeError, ImportError):
            # We expect PPO1 load to fail if mpi4py isn't installed.
            # Stable Baselines can be installed without mpi4py.
            tf.logging.debug(f"Couldn't load {cls_name}. Skipping...")


STABLE_BASELINES_CLASSES = {
    "ppo1": ("stable_baselines:PPO1", "policy_pi"),
    "ppo2": ("stable_baselines:PPO2", "act_model"),
}
_add_stable_baselines_policies(STABLE_BASELINES_CLASSES)


def load_policy(
    policy_type: str, policy_path: str, venv: VecEnv
) -> ContextManager[BasePolicy]:
    """Load serialized policy.

    Args:
        policy_type: A key in `policy_registry`, e.g. `ppo2`.
        policy_path: A path on disk where the policy is stored.
        venv: An environment that the policy is to be used with.
    """
    agent_loader = policy_registry.get(policy_type)
    return agent_loader(policy_path, venv)


def save_stable_model(
    output_dir: str,
    model: BaseRLModel,
    vec_normalize: Optional[VecNormalize] = None,
) -> None:
    """Serialize policy.

    Load later with `load_policy(..., policy_path=output_dir)`.

    Args:
        output_dir: Path to the save directory.
        policy: The stable baselines policy.
        vec_normalize: Optionally, a VecNormalize to save statistics for.
            `load_policy` automatically applies `NormalizePolicy` wrapper
            when loading.
    """
    os.makedirs(output_dir, exist_ok=True)
    model.save(os.path.join(output_dir, "model.pkl"))
    if vec_normalize is not None:
        with open(os.path.join(output_dir, "vec_normalize.pkl"), "wb") as f:
            pickle.dump(vec_normalize, f)
    tf.logging.info("Saved policy to %s", output_dir)



import os

import sacred

from imitation.scripts.config.common import DEFAULT_INIT_RL_KWARGS
from imitation.util import util

expert_demos_ex = sacred.Experiment("expert_demos")


@expert_demos_ex.config
def expert_demos_defaults():
    env_name = "CartPole-v1"  # The gym.Env name
    total_timesteps = int(1e6)  # Number of training timesteps in model.learn()
    num_vec = 8  # Number of environments in VecEnv
    parallel = True  # Use SubprocVecEnv (generally faster if num_vec>1)
    normalize = True  # Use VecNormalize
    normalize_kwargs = dict()  # kwargs for `VecNormalize`
    max_episode_steps = None  # Set to positive int to limit episode horizons
    n_episodes_eval = 50  # Num of episodes for final ep reward mean evaluation

    init_rl_kwargs = dict(DEFAULT_INIT_RL_KWARGS)

    # If specified, overrides the ground-truth environment reward
    reward_type = None  # override reward type
    reward_path = None  # override reward path

    rollout_save_interval = -1  # Num updates between saves (<=0 disables)
    rollout_save_final = True  # If True, save after training is finished.
    rollout_save_n_timesteps = None  # Min timesteps saved per file, optional.
    rollout_save_n_episodes = None  # Num episodes saved per file, optional.

    policy_save_interval = 100  # Num updates between saves (<=0 disables)
    policy_save_final = True  # If True, save after training is finished.

    init_tensorboard = False  # If True, then write Tensorboard logs.

    log_root = os.path.join("output", "expert_demos")  # output directory


@expert_demos_ex.config
def default_end_cond(rollout_save_n_timesteps, rollout_save_n_episodes):
    # Only set default if both end cond options are None.
    # This way the Sacred CLI caller can set `rollout_save_n_episodes` only
    # without getting an error that `rollout_save_n_timesteps is not None`.
    if rollout_save_n_timesteps is None and rollout_save_n_episodes is None:
        rollout_save_n_timesteps = 2000  # Min timesteps saved per file, optional.


@expert_demos_ex.config
def logging(env_name, log_root):
    log_dir = os.path.join(
        log_root, env_name.replace("/", "_"), util.make_unique_timestamp()
    )


@expert_demos_ex.config
def rollouts_from_policy_only_defaults(log_dir):
    policy_path = None  # Policy path for rollouts_from_policy command only
    policy_type = "ppo2"  # Policy type for rollouts_from_policy command only
    rollout_save_path = os.path.join(
        log_dir, "rollout.pkl"
    )  # Save path for `rollouts_from_policy` only.


# Standard Gym env configs


@expert_demos_ex.named_config
def acrobot():
    env_name = "Acrobot-v1"


@expert_demos_ex.named_config
def ant():
    env_name = "Ant-v2"
    locals().update(**ant_shared_locals)


@expert_demos_ex.named_config
def cartpole():
    env_name = "CartPole-v1"
    total_timesteps = int(1e5)


@expert_demos_ex.named_config
def half_cheetah():
    env_name = "HalfCheetah-v2"
    total_timesteps = int(5e6)  # does OK after 1e6, but continues improving


@expert_demos_ex.named_config
def hopper():
    # TODO(adam): upgrade to Hopper-v3?
    env_name = "Hopper-v2"


@expert_demos_ex.named_config
def humanoid():
    env_name = "Humanoid-v2"
    init_rl_kwargs = dict(
        n_steps=2048,
    )  # batch size of 2048*8=16384 due to num_vec
    total_timesteps = int(10e6)  # fairly discontinuous, needs at least 5e6


@expert_demos_ex.named_config
def mountain_car():
    env_name = "MountainCar-v0"


@expert_demos_ex.named_config
def pendulum():
    env_name = "Pendulum-v0"


@expert_demos_ex.named_config
def reacher():
    env_name = "Reacher-v2"


@expert_demos_ex.named_config
def swimmer():
    env_name = "Swimmer-v2"


@expert_demos_ex.named_config
def walker():
    env_name = "Walker2d-v2"


# Custom env configs


@expert_demos_ex.named_config
def custom_ant():
    env_name = "imitation/CustomAnt-v0"
    locals().update(**ant_shared_locals)


@expert_demos_ex.named_config
def disabled_ant():
    env_name = "imitation/DisabledAnt-v0"
    locals().update(**ant_shared_locals)


@expert_demos_ex.named_config
def two_d_maze():
    env_name = "imitation/TwoDMaze-v0"


# Debug configs


@expert_demos_ex.named_config
def fast():
    """Intended for testing purposes: small # of updates, ends quickly."""
    total_timesteps = int(1e3)
    max_episode_steps = int(1e3)


# Shared settings

ant_shared_locals = dict(
    init_rl_kwargs=dict(
        n_steps=2048,
    ),  # batch size of 2048*8=16384 due to num_vec
    total_timesteps=int(5e6),
    max_episode_steps=500,  # To match `inverse_rl` settings.
)



"""Configuration for imitation.scripts.train_adversarial."""

import os

import sacred
from stable_baselines.common import policies

from imitation.policies import base
from imitation.scripts.config.common import DEFAULT_INIT_RL_KWARGS
from imitation.util import util

train_ex = sacred.Experiment("train_adversarial", interactive=True)


@train_ex.config
def train_defaults():
    env_name = "CartPole-v1"  # environment to train on
    total_timesteps = 1e5  # Num of environment transitions to sample
    algorithm = "gail"  # Either "airl" or "gail"

    n_expert_demos = None  # Num demos used. None uses every demo possible
    n_episodes_eval = 50  # Num of episodes for final mean ground truth return

    # Number of epochs in between plots (<0 disables) (=0 means final plot only)
    plot_interval = -1
    n_plot_episodes = 5  # Number of rollouts for each mean_ep_rew data
    # Interval for extra episode rew data. (<=0 disables)
    extra_episode_data_interval = -1
    show_plots = True  # Show plots in addition to saving them

    # Number of environments in VecEnv, must evenly divide gen_batch_size
    num_vec = 8

    # Use SubprocVecEnv rather than DummyVecEnv (generally faster if num_vec>1)
    parallel = True
    max_episode_steps = None  # Set to positive int to limit episode horizons

    # Kwargs for initializing GAIL and AIRL
    algorithm_kwargs = dict(
        shared=dict(
            disc_batch_size=2048,  # Batch size for discriminator updates
            disc_minibatch_size=512,  # Num discriminator updates per batch
        ),
        airl={},
        gail={},
    )

    # Kwargs for initializing {GAIL,AIRL}DiscrimNet
    discrim_net_kwargs = dict(shared={}, airl={}, gail={})

    # Modifies the __init__ arguments for the imitation policy
    init_rl_kwargs = dict(
        policy_class=base.FeedForward32Policy,
        **DEFAULT_INIT_RL_KWARGS,
    )
    gen_batch_size = 2048  # Batch size for generator updates

    log_root = os.path.join("output", "train_adversarial")  # output directory
    checkpoint_interval = 0  # Num epochs between checkpoints (<0 disables)
    init_tensorboard = False  # If True, then write Tensorboard logs
    rollout_hint = None  # Used to generate default rollout_path
    data_dir = "data/"  # Default data directory


@train_ex.config
def aliases_default_gen_batch_size(algorithm_kwargs, gen_batch_size):
    # Setting generator buffer capacity and discriminator batch size to
    # the same number is equivalent to not using a replay buffer at all.
    # "Disabling" the replay buffer seems to improve convergence speed, but may
    # come at a cost of stability.

    algorithm_kwargs["shared"]["gen_replay_buffer_capacity"] = gen_batch_size


@train_ex.config
def calc_n_steps(num_vec, gen_batch_size):
    assert gen_batch_size % num_vec == 0, "num_vec must evenly divide gen_batch_size"
    init_rl_kwargs = dict(n_steps=gen_batch_size // num_vec)


@train_ex.config
def paths(env_name, log_root, rollout_hint, data_dir):
    log_dir = os.path.join(
        log_root, env_name.replace("/", "_"), util.make_unique_timestamp()
    )

    # Recommended that user sets rollout_path manually.
    # By default we guess the named config associated with `env_name`
    # and attempt to load rollouts from `data/expert_models/`.
    if rollout_hint is None:
        rollout_hint = env_name.split("-")[0].lower()
    rollout_path = os.path.join(
        data_dir, "expert_models", f"{rollout_hint}_0", "rollouts", "final.pkl"
    )


# Training algorithm named configs


@train_ex.named_config
def gail():
    """Quick alias for algorithm=gail"""
    algorithm = "gail"


@train_ex.named_config
def airl():
    """Quick alias for algorithm=airl"""
    algorithm = "airl"


# Shared settings

MUJOCO_SHARED_LOCALS = dict(discrim_net_kwargs=dict(airl=dict(entropy_weight=0.1)))

ANT_SHARED_LOCALS = dict(
    total_timesteps=3e7,
    max_episode_steps=500,  # To match `inverse_rl` settings.
    algorithm_kwargs=dict(shared=dict(disc_batch_size=2048 * 8)),
    gen_batch_size=2048 * 8,
)


# Classic RL Gym environment named configs


@train_ex.named_config
def acrobot():
    env_name = "Acrobot-v1"
    rollout_hint = "acrobot"


@train_ex.named_config
def cartpole():
    env_name = "CartPole-v1"
    rollout_hint = "cartpole"
    discrim_net_kwargs = {"gail": {"scale": False}}


@train_ex.named_config
def mountain_car():
    env_name = "MountainCar-v0"
    rollout_hint = "mountain_car"


@train_ex.named_config
def pendulum():
    env_name = "Pendulum-v0"
    rollout_hint = "pendulum"


# Standard MuJoCo Gym environment named configs


@train_ex.named_config
def ant():
    locals().update(**MUJOCO_SHARED_LOCALS)
    locals().update(**ANT_SHARED_LOCALS)
    env_name = "Ant-v2"
    rollout_hint = "ant"


@train_ex.named_config
def half_cheetah():
    locals().update(**MUJOCO_SHARED_LOCALS)
    env_name = "HalfCheetah-v2"
    rollout_hint = "half_cheetah"
    total_timesteps = 2e6


@train_ex.named_config
def hopper():
    locals().update(**MUJOCO_SHARED_LOCALS)
    # TODO(adam): upgrade to Hopper-v3?
    env_name = "Hopper-v2"
    rollout_hint = "hopper"


@train_ex.named_config
def humanoid():
    locals().update(**MUJOCO_SHARED_LOCALS)
    env_name = "Humanoid-v2"
    rollout_hint = "humanoid"
    total_timesteps = 4e6


@train_ex.named_config
def reacher():
    env_name = "Reacher-v2"
    rollout_hint = "reacher"


@train_ex.named_config
def swimmer():
    locals().update(**MUJOCO_SHARED_LOCALS)
    env_name = "Swimmer-v2"
    rollout_hint = "swimmer"
    total_timesteps = 2e6
    init_rl_kwargs = dict(policy_network_class=policies.MlpPolicy)


@train_ex.named_config
def walker():
    locals().update(**MUJOCO_SHARED_LOCALS)
    env_name = "Walker2d-v2"
    rollout_hint = "walker"


# Custom Gym environment named configs


@train_ex.named_config
def two_d_maze():
    locals().update(**MUJOCO_SHARED_LOCALS)
    env_name = "imitation/TwoDMaze-v0"
    rollout_hint = "two_d_maze"


@train_ex.named_config
def custom_ant():
    locals().update(**MUJOCO_SHARED_LOCALS)
    # Watch out -- ANT_SHARED_LOCALS could erroneously erase nested dict keys from
    # MUJOCO_SHARED_LOCALS because `locals().update()` doesn't merge dicts
    # "Sacred-style".
    locals().update(**ANT_SHARED_LOCALS)
    env_name = "imitation/CustomAnt-v0"
    rollout_hint = "custom_ant"


@train_ex.named_config
def disabled_ant():
    locals().update(**MUJOCO_SHARED_LOCALS)
    locals().update(**ANT_SHARED_LOCALS)
    env_name = "imitation/DisabledAnt-v0"
    rollout_hint = "disabled_ant"


# Debug configs


@train_ex.named_config
def fast():
    """Minimize the amount of computation.

    Useful for test cases.
    """
    total_timesteps = 10
    n_expert_demos = 1
    n_episodes_eval = 1
    algorithm_kwargs = dict(shared=dict(disc_batch_size=2, disc_minibatch_size=2))
    gen_batch_size = 2
    parallel = False  # easier to debug with everything in one process
    max_episode_steps = 100
    num_vec = 2
    init_rl_kwargs = dict(nminibatches=1)



import collections
import os
import pickle
from abc import ABC, abstractmethod
from typing import Callable, Dict, Optional, Sequence, Tuple

import gym
import numpy as np
import tensorflow as tf

from imitation.rewards import reward_net
from imitation.util import networks, serialize


class DiscrimNet(serialize.Serializable, ABC):
    """Abstract base class for discriminator, used in AIRL and GAIL."""

    def __init__(self):
        self._sess = tf.get_default_session()

        # Dict from names to scalar Tensors that are logged by
        # `AdversarialTrainer.train_disc_step`.
        self._train_stats = collections.OrderedDict()  # type: Dict[str, tf.Tensor]

        # Build necessary placeholders, then construct rest of the graph.
        # _labels_gen_is_one_ph holds the label of every state-action pair that the
        # discriminator is being trained on. Use 0.0 for expert policy. Use 1.0 for
        # generated policy.
        self._labels_gen_is_one_ph = tf.placeholder(
            shape=(None,), dtype=tf.int32, name="discrim_labels_gen_is_one"
        )
        # This placeholder holds the generator-policy log action probabilities,
        # $\log \pi(a \mid s)$, of each state-action pair. This includes both
        # actions taken by the generator *and* those by the expert (we can
        # ask our policy for action probabilities even if we don't take them).
        # TODO(adam): this is only used by AIRL; sad we have to always include it
        self._log_policy_act_prob_ph = tf.placeholder(
            shape=(None,), dtype=tf.float32, name="log_ro_act_prob_ph"
        )

        self._disc_loss = None  # type: tf.Tensor
        self._policy_train_reward = None  # type: tf.Tensor
        self._policy_test_reward = None  # type: tf.Tensor
        self._disc_logits_gen_is_high = None  # type: tf.Tensor

        self.build_graph()

        assert self._disc_loss is not None
        assert self._policy_train_reward is not None
        assert self._policy_test_reward is not None
        assert self._disc_logits_gen_is_high is not None

        self._add_default_train_stats()
        tf.summary.histogram("disc_logits", self.disc_logits_gen_is_high)

    @property
    @abstractmethod
    def obs_ph(self):
        """The previous observation placeholder."""

    @property
    @abstractmethod
    def act_ph(self):
        """The action placeholder."""

    @property
    @abstractmethod
    def next_obs_ph(self):
        """The new observation placeholder."""

    @property
    @abstractmethod
    def done_ph(self):
        """The episode termination placeholder."""

    @property
    def labels_gen_is_one_ph(self):
        """The expert (0.0) or generated (1.0) labels placeholder."""
        return self._labels_gen_is_one_ph

    @property
    def log_policy_act_prob_ph(self):
        """The log-probability of policy actions placeholder."""
        return self._log_policy_act_prob_ph

    @property
    def disc_logits_gen_is_high(self) -> tf.Tensor:
        """The discriminator's logits for each state-action sample.

        A high value corresponds to predicting generator, and a low value corresponds to
        predicting expert.
        """
        return self._disc_logits_gen_is_high

    @property
    def disc_loss(self) -> tf.Tensor:
        return self._disc_loss  # pytype: disable=attribute-error

    @property
    def policy_train_reward(self) -> tf.Tensor:
        return self._policy_train_reward  # pytype: disable=attribute-error

    @property
    def policy_test_reward(self) -> tf.Tensor:
        return self._policy_test_reward  # pytype: disable=attribute-error

    @property
    def train_stats(self) -> Dict[str, tf.Tensor]:
        """A feed dictionary of scalar Tensors to be logged during training."""
        return self._train_stats

    @abstractmethod
    def build_graph(self):
        """Builds forward prop graph, reward, loss, and summary ops.

        Gets called once during construction. Should create the following
        attributes:

        - `self._policy_train_reward`: reward used during imitation learning,
          should be a rank-1 Tensor.
        - `self._policy_test_reward`: reward used during testing, should be rank-1
          Tensor. This is useful for AIRL, where it can be used to drop shaping
          terms, but for GAIL it will just be an alias for _policy_train_reward.
        - `self._disc_loss`: discriminator loss to be optimised.

        FIXME(sam): remove train reward/test reward distinction.
        """

    def _add_default_train_stats(self) -> None:
        """Updates `self._train_stats` with several default scalar logging Tensors."""
        bin_is_generated_pred = self.disc_logits_gen_is_high > 0
        bin_is_generated_true = self.labels_gen_is_one_ph > 0
        bin_is_expert_true = tf.logical_not(bin_is_generated_true)
        int_is_generated_pred = tf.cast(bin_is_generated_pred, tf.int32)
        int_is_generated_true = tf.cast(bin_is_generated_true, tf.int32)
        n_generated = tf.reduce_sum(int_is_generated_true)
        n_labels = tf.size(self.labels_gen_is_one_ph)
        n_expert = n_labels - n_generated
        pct_expert = tf.cast(n_expert, tf.float32) / tf.cast(n_labels, tf.float32)
        n_expert_pred = tf.size(bin_is_generated_pred) - tf.reduce_sum(
            int_is_generated_pred
        )
        pct_expert_pred = tf.cast(n_expert_pred, tf.float32) / tf.cast(
            n_labels, tf.float32
        )
        correct_vec = tf.equal(bin_is_generated_pred, bin_is_generated_true)
        acc = tf.reduce_mean(tf.cast(correct_vec, tf.float32))

        _n_pred_expert = tf.reduce_sum(
            tf.cast(tf.logical_and(bin_is_expert_true, correct_vec), tf.float32)
        )
        _n_expert_or_1 = tf.cast(tf.maximum(1, n_expert), tf.float32)
        expert_acc = _n_pred_expert / _n_expert_or_1

        _n_pred_gen = tf.reduce_sum(
            tf.cast(tf.logical_and(bin_is_generated_true, correct_vec), tf.float32)
        )
        _n_gen_or_1 = tf.cast(tf.maximum(1, n_generated), tf.float32)
        generated_acc = _n_pred_gen / _n_gen_or_1

        label_dist = tf.distributions.Bernoulli(logits=self.disc_logits_gen_is_high)
        entropy = tf.reduce_mean(label_dist.entropy())

        self._train_stats.update(
            [
                ("disc_xent_loss", tf.reduce_mean(self._disc_loss)),
                # accuracy, as well as accuracy on *just* expert examples and *just*
                # generated examples
                ("disc_acc", acc),
                ("disc_acc_expert", expert_acc),
                ("disc_acc_gen", generated_acc),
                # entropy of the predicted label distribution, averaged equally across
                # both classes (if this drops then disc is very good or has given up)
                ("disc_entropy", entropy),
                # true number of expert demos and predicted number of expert demos
                ("disc_proportion_expert_true", pct_expert),
                ("disc_proportion_expert_pred", pct_expert_pred),
                ("n_expert", n_expert),
                ("n_generated", n_generated),
            ]
        )

    def reward_train(
        self,
        obs: np.ndarray,
        act: np.ndarray,
        next_obs: np.ndarray,
        dones: np.ndarray,
        *,
        gen_log_prob_fn: Callable[..., np.ndarray],
    ) -> np.ndarray:
        """Vectorized reward for training an imitation learning algorithm.

        Args:
            obs: The observation input. Its shape is
                `(batch_size,) + observation_space.shape`.
            act: The action input. Its shape is
                `(batch_size,) + action_space.shape`. The None dimension is
                expected to be the same as None dimension from `obs_input`.
            next_obs: The observation input. Its shape is
                `(batch_size,) + observation_space.shape`.
            dones: Whether the episode has terminated. Its shape is `(batch_size,)`.
            gen_log_prob_fn: The generator policy's action probabilities function.
                A Callable such that
                `log_act_prob_fn(observations=obs, actions=act, lopg=True)`
                returns `log_act_prob`, the generator's log action probabilities.
                `log_act_prob[i]` is equal to the generator's log probability of
                choosing `act[i]` given `obs[i]`.
        Returns:
            The rewards. Its shape is `(batch_size,)`.
        """
        log_act_prob = gen_log_prob_fn(
            observation=obs, actions=act, logp=True
        ).flatten()

        n_gen = len(obs)
        assert obs.shape == next_obs.shape
        assert len(act) == n_gen
        assert log_act_prob.shape == (n_gen,)

        fd = {
            self.obs_ph: obs,
            self.act_ph: act,
            self.next_obs_ph: next_obs,
            self.done_ph: dones,
            self.labels_gen_is_one_ph: np.ones(n_gen),
            self.log_policy_act_prob_ph: log_act_prob,
        }
        rew = self._sess.run(self.policy_train_reward, feed_dict=fd)
        assert rew.shape == (n_gen,)
        return rew

    def reward_test(
        self,
        obs: np.ndarray,
        act: np.ndarray,
        next_obs: np.ndarray,
        dones: np.ndarray,
    ) -> np.ndarray:
        """Vectorized reward for training an expert during transfer learning.

        Args:
            obs: The observation input. Its shape is
                `(batch_size,) + observation_space.shape`.
            act: The action input. Its shape is
                `(batch_size,) + action_space.shape`. The None dimension is
                expected to be the same as None dimension from `obs_input`.
            next_obs: The observation input. Its shape is
                `(batch_size,) + observation_space.shape`.
            dones: Whether the episode has terminated. Its shape is `(batch_size,)`.
        Returns:
            The rewards. Its shape is `(batch_size,)`.
        """
        fd = {
            self.obs_ph: obs,
            self.act_ph: act,
            self.next_obs_ph: next_obs,
            self.done_ph: dones,
        }
        rew = self._sess.run(self.policy_test_reward, feed_dict=fd)
        assert rew.shape == (len(obs),)
        return rew


class DiscrimNetAIRL(DiscrimNet):
    r"""The AIRL discriminator for a given RewardNet.

    The AIRL discriminator is of the form
    .. math:: D_{\theta}(s,a) = \frac{\exp(f_{\theta}(s,a)}{\exp(f_{\theta}(s,a) + \pi(a \mid s)}

    where :math:`f_{\theta}` is `self.reward_net`.
    """  # noqa: E501

    def __init__(self, reward_net: reward_net.RewardNet, entropy_weight: float = 1.0):
        """Builds a DiscrimNetAIRL.

        Args:
            reward_net: A RewardNet, used as $f_{\theta}$ in the discriminator.
            entropy_weight: The coefficient for the entropy regularization term.
                To match the AIRL derivation, it should be 1.0.
                However, empirically a lower value sometimes work better.
        """
        self.reward_net = reward_net
        self.entropy_weight = entropy_weight
        super().__init__()
        tf.logging.info("Using AIRL")

    @property
    def obs_ph(self):
        return self.reward_net.obs_ph

    @property
    def act_ph(self):
        return self.reward_net.act_ph

    @property
    def next_obs_ph(self):
        return self.reward_net.next_obs_ph

    @property
    def done_ph(self):
        return self.reward_net.done_ph

    def build_graph(self):
        self._disc_logits_gen_is_high = (
            self.log_policy_act_prob_ph - self.reward_net.reward_output_train
        )

        # Construct discriminator loss.
        self._disc_loss = tf.nn.sigmoid_cross_entropy_with_logits(
            labels=tf.cast(self.labels_gen_is_one_ph, tf.float32),
            logits=self._disc_logits_gen_is_high,
            name="disc_loss",
        )  # (None,)

        # Construct generator reward:
        # \[\hat{r}(s,a) = \log(D_{\theta}(s,a)) - \log(1 - D_{\theta}(s,a)).\]
        # This simplifies to:
        # \[\hat{r}(s,a) = f_{\theta}(s,a) - \log \pi(a \mid s).\]
        # This is just an entropy-regularized objective.

        ent_bonus = -self.entropy_weight * self.log_policy_act_prob_ph
        self._policy_train_reward = self.reward_net.reward_output_train + ent_bonus
        self._policy_test_reward = self.reward_net.reward_output_test

        self.reward_net.build_summaries()

    def _save(self, directory):
        os.makedirs(directory, exist_ok=True)
        params = {
            "entropy_weight": self.entropy_weight,
            "reward_net_cls": type(self.reward_net),
        }
        with open(os.path.join(directory, "args"), "wb") as f:
            pickle.dump(params, f)
        return self.reward_net.save(os.path.join(directory, "reward_net"))

    @classmethod
    def _load(cls, directory):
        with open(os.path.join(directory, "args"), "rb") as f:
            params = pickle.load(f)
        reward_net_cls = params.pop("reward_net_cls")
        reward_net = reward_net_cls.load(os.path.join(directory, "reward_net"))
        return cls(reward_net=reward_net, **params)


DiscrimNetBuilder = Callable[
    [Sequence[tf.Tensor]], Tuple[tf.Tensor, networks.LayersDict]
]
"""Type alias for function that builds a discriminator network.

Takes an observation and action tensor and produces a tuple containing
(1) a list of used TF layers, and (2) output logits.
"""


class DiscrimNetGAIL(DiscrimNet, serialize.LayersSerializable):
    """The discriminator to use for GAIL."""

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        build_discrim_net: DiscrimNetBuilder = networks.build_and_apply_mlp,
        build_discrim_net_kwargs: Optional[dict] = None,
        scale: bool = False,
    ):
        """Construct discriminator network.

        Args:
          observation_space: observation space for this environment.
          action_space: action space for this environment:
          build_discrim_net: a callable that takes an observation input tensor
            and action input tensor as input, then computes the logits
            necessary to feed to GAIL. When called, the function should return
            *both* a `LayersDict` containing all the layers used in
            construction of the discriminator network, and a `tf.Tensor`
            representing the desired discriminator logits.
          build_discrim_net_kwargs: optional extra keyword arguments for
            `build_discrim_net()`.
          scale: should inputs be rescaled according to declared observation
            space bounds?
        """
        # for serialisation
        args = dict(locals())

        # things we'll need in .build_graph()
        self._observation_space = observation_space
        self._action_space = action_space
        self._scale = scale
        self._build_discrim_net = build_discrim_net
        self._build_discrim_net_kwargs = build_discrim_net_kwargs or {}

        self._disc_mlp = None
        # Builds graph via call to `self.build_graph()`.
        DiscrimNet.__init__(self)
        assert self._disc_mlp is not None

        # records args for un-pickling as well as newly-created model
        serialize.LayersSerializable.__init__(**args, layers=self._disc_mlp)

        tf.logging.info("using GAIL")

    @property
    def obs_ph(self):
        return self._obs_ph

    @property
    def act_ph(self):
        return self._act_ph

    @property
    def next_obs_ph(self):
        return self._next_obs_ph

    @property
    def done_ph(self):
        return self._done_ph

    def build_graph(self):
        phs, inps = networks.build_inputs(
            self._observation_space, self._action_space, scale=self._scale
        )
        self._obs_ph, self._act_ph, self._next_obs_ph, self._done_ph = phs
        self.obs_input, self.act_input, _, self.done_input = inps

        with tf.variable_scope("discrim_network"):
            self._disc_logits_gen_is_high, self._disc_mlp = self._build_discrim_net(
                [self.obs_input, self.act_input], **self._build_discrim_net_kwargs
            )
        self._policy_test_reward = self._policy_train_reward = -tf.log_sigmoid(
            self._disc_logits_gen_is_high
        )

        self._disc_loss = tf.nn.sigmoid_cross_entropy_with_logits(
            logits=self._disc_logits_gen_is_high,
            labels=tf.cast(self.labels_gen_is_one_ph, tf.float32),
        )



import abc
import dataclasses
from typing import Dict, Generic, Mapping, Optional, Tuple, Type, TypeVar

import numpy as np

from imitation.data import types

T = TypeVar("T")


class Dataset(abc.ABC, Generic[T]):
    @abc.abstractmethod
    def sample(self: "Dataset[T]", n_samples: int) -> T:
        """Return a batch of data.

        Args:
            n_samples: A positive integer indicating the number of samples to return.
        Raises:
            ValueError: If n_samples is nonpositive.
        """

    @abc.abstractmethod
    def size(self: "Dataset[T]") -> Optional[int]:
        """Returns the number of samples in this dataset, ie the epoch size.

        Returns None if not known or undefined.
        """


class DictDataset(Dataset[Dict[str, np.ndarray]]):
    def __init__(self, data_map: Mapping[str, np.ndarray]):
        """Abstract base class for sampling data from an in-memory dictionary.

        The return value of `.sample(n_samples)` is a dictionary with the same keys as
        `data_map` and whose values are `n_samples` stacked rows selected from the
        values of `data_map`. In the return value, rows that were previously parallel
        remain parallel. As an example, if
        `data_map=dict(a=np.arange(5), b=(np.arange(5) + 4))`, then it is guaranteed
        that every call to `.sample()` returns a dictionary `batch` such that
        `batch['a'] == batch['b'] + 4`.

        Args:
            data_map: A mapping from keys to np.ndarray values, where every value has
                an equal number of rows. This argument is not never mutated, as each
                array is immediately copied. Required to be non-empty.
        """
        if len(data_map) == 0:
            raise ValueError("Empty data_map not allowed.")
        self.data_map = {k: v.copy() for k, v in data_map.items()}

        n_rows_set = set(len(v) for v in data_map.values())
        if len(n_rows_set) != 1:
            raise ValueError(f"Unequal number of rows in data_map values: {n_rows_set}")
        self._n_data = next(iter(n_rows_set))

    def size(self):
        return self._n_data


class EpochOrderDictDataset(DictDataset):
    def __init__(self, data_map: Mapping[str, np.ndarray], shuffle: bool = True):
        """In-memory dict data sampler that samples in epoch-order.

        No sample from `data_map` can be returned an X+1th time by `sample()` until
        every other sample has been returned X times.

        Args:
            data_map: A mapping from keys to np.ndarray values, where every value has
                an equal number of rows. This argument is not mutated, as each array
                is immediately copied.
            shuffle: If true, then shuffle the dataset upon initialization and at the
                end of every epoch.
        """
        super().__init__(data_map)
        self._next_idx = 0
        self._shuffle = shuffle
        if shuffle:
            self.shuffle_dataset()

    def shuffle_dataset(self) -> None:
        """Shuffles the data_map in place."""
        perm = np.arange(self.size())
        np.random.shuffle(perm)
        for key in self.data_map:
            self.data_map[key] = self.data_map[key][perm]

    def sample(self, n_samples: int) -> Dict[str, np.ndarray]:
        if n_samples <= 0:
            raise ValueError(f"n_samples={n_samples} must be positive.")
        samples_accum = {k: [] for k in self.data_map.keys()}
        n_samples_remain = n_samples
        while n_samples_remain > 0:
            n_samples_actual, sample = self._sample_bounded(n_samples_remain)
            n_samples_remain -= n_samples_actual
            for k, v in sample.items():
                samples_accum[k].append(v)

        result = {k: np.copy(np.concatenate(v)) for k, v in samples_accum.items()}
        assert all(len(v) == n_samples for v in result.values())
        return result

    def _sample_bounded(
        self, n_samples_request: int
    ) -> Tuple[int, Dict[str, np.ndarray]]:
        """Like `.sample()`, but allowed to return fewer samples on epoch boundaries."""
        assert n_samples_request > 0
        if self._next_idx >= self.size():
            self._next_idx = 0
            if self._shuffle:
                self.shuffle_dataset()

        n_samples_actual = min(n_samples_request, self.size() - self._next_idx)
        assert n_samples_actual > 0

        result = {}
        for key in self.data_map:
            result[key] = self.data_map[key][
                self._next_idx : self._next_idx + n_samples_actual
            ]
            assert len(result[key]) == n_samples_actual
        self._next_idx += n_samples_actual
        return n_samples_actual, result


class RandomDictDataset(DictDataset):
    """In-memory data sampler that uniformly samples with replacement."""

    def sample(self, n_samples: int) -> Dict[str, np.ndarray]:
        if n_samples <= 0:
            raise ValueError(f"n_samples={n_samples} must be positive.")
        inds = np.random.randint(self.size(), size=n_samples)
        return {k: v[inds] for k, v in self.data_map.items()}


S = TypeVar("S", bound=types.TransitionsMinimal)


class TransitionsDictDatasetAdaptor(Dataset[S]):
    def __init__(
        self,
        transitions: S,
        dict_dataset_cls: Type[DictDataset] = RandomDictDataset,
        dict_dataset_cls_kwargs: Optional[Mapping] = None,
    ):
        """Adapts a `DictDataset` class to build `Transitions` `Dataset`.

        Args:
            transitions: An `TransitionsMinimal` instance to be sampled from and stored
                internally as a `dict` of copied arrays. Note that `sample` will return
                the same subclass type of `TransitionsMinimal` as the type of this arg.
            dict_dataset_cls: `DictDataset` class to be adapted.
            dict_dataset_kwargs: Optional kwargs for initializing `DictDataset` class.
        """
        data_map: Dict[str, np.ndarray] = dataclasses.asdict(transitions)
        kwargs = dict_dataset_cls_kwargs or {}
        self.transitions_cls: Type[S] = type(transitions)
        self.simple_dataset = dict_dataset_cls(
            data_map, **kwargs  # pytype: disable=not-instantiable
        )

    def sample(self, n_samples: int) -> S:
        dict_samples = self.simple_dataset.sample(n_samples)
        result = self.transitions_cls(**dict_samples)
        assert len(result) == n_samples
        return result

    def size(self):
        return self.simple_dataset.size()



import contextlib
import os
import os.path as osp
import time
from typing import Optional

import tensorflow as tf
from sacred.observers import FileStorageObserver
from stable_baselines.common.vec_env import VecEnvWrapper

import imitation.util.sacred as sacred_util
from imitation.data import rollout
from imitation.policies import serialize
from imitation.rewards.serialize import load_reward
from imitation.scripts.config.eval_policy import eval_policy_ex
from imitation.util import reward_wrapper, util


class InteractiveRender(VecEnvWrapper):
    def __init__(self, venv, fps):
        super().__init__(venv)
        self.render_fps = fps

    def reset(self):
        ob = self.venv.reset()
        self.venv.render()
        return ob

    def step_wait(self):
        ob = self.venv.step_wait()
        if self.render_fps > 0:
            time.sleep(1 / self.render_fps)
        self.venv.render()
        return ob


@eval_policy_ex.main
def eval_policy(
    _run,
    _seed: int,
    env_name: str,
    eval_n_timesteps: Optional[int],
    eval_n_episodes: Optional[int],
    num_vec: int,
    parallel: bool,
    render: bool,
    render_fps: int,
    log_dir: str,
    policy_type: str,
    policy_path: str,
    reward_type: Optional[str] = None,
    reward_path: Optional[str] = None,
    max_episode_steps: Optional[int] = None,
):
    """Rolls a policy out in an environment, collecting statistics.

    Args:
      _seed: generated by Sacred.
      env_name: Gym environment identifier.
      eval_n_timesteps: Minimum number of timesteps to evaluate for. Set exactly
          one of `eval_n_episodes` and `eval_n_timesteps`.
      eval_n_episodes: Minimum number of episodes to evaluate for. Set exactly
          one of `eval_n_episodes` and `eval_n_timesteps`.
      num_vec: Number of environments to run simultaneously.
      parallel: If True, use `SubprocVecEnv` for true parallelism; otherwise,
          uses `DummyVecEnv`.
      max_episode_steps: If not None, then environments are wrapped by
          TimeLimit so that they have at most `max_episode_steps` steps per
          episode.
      render: If True, renders interactively to the screen.
      log_dir: The directory to log intermediate output to. (As of 2019-07-19
          this is just episode-by-episode reward from bench.Monitor.)
      policy_type: A unique identifier for the saved policy,
          defined in POLICY_CLASSES.
      policy_path: A path to the serialized policy.
      reward_type: If specified, overrides the environment reward with
          a reward of this.
      reward_path: If reward_type is specified, the path to a serialized reward
          of `reward_type` to override the environment reward with.

    Returns:
      Return value of `imitation.util.rollout.rollout_stats()`.
    """
    os.makedirs(log_dir, exist_ok=True)
    sacred_util.build_sacred_symlink(log_dir, _run)

    tf.logging.set_verbosity(tf.logging.INFO)
    tf.logging.info("Logging to %s", log_dir)
    sample_until = rollout.make_sample_until(eval_n_timesteps, eval_n_episodes)
    venv = util.make_vec_env(
        env_name,
        num_vec,
        seed=_seed,
        parallel=parallel,
        log_dir=log_dir,
        max_episode_steps=max_episode_steps,
    )

    if render:
        venv = InteractiveRender(venv, render_fps)
    # TODO(adam): add support for videos using VideoRecorder?

    with contextlib.ExitStack() as stack:
        if reward_type is not None:
            reward_fn_ctx = load_reward(reward_type, reward_path, venv)
            reward_fn = stack.enter_context(reward_fn_ctx)
            venv = reward_wrapper.RewardVecEnvWrapper(venv, reward_fn)
            tf.logging.info(f"Wrapped env in reward {reward_type} from {reward_path}.")

        with serialize.load_policy(policy_type, policy_path, venv) as policy:
            trajs = rollout.generate_trajectories(policy, venv, sample_until)
    return rollout.rollout_stats(trajs)


def main_console():
    observer = FileStorageObserver(osp.join("output", "sacred", "eval_policy"))
    eval_policy_ex.observers.append(observer)
    eval_policy_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



import numpy as np
from gym import utils

from imitation.envs.examples.airl_envs.common import get_asset_xml
from imitation.envs.examples.airl_envs.twod_mjc_env import TwoDEnv

INIT_POS = np.array([0.15, 0.15])
TARGET = np.array([0.15, -0.15])
DIST_THRESH = 0.12


class TwoDMaze(TwoDEnv, utils.EzPickle):
    def __init__(self, verbose=False):
        self.verbose = verbose
        self.max_episode_length = 200
        self.episode_length = 0
        utils.EzPickle.__init__(self)
        TwoDEnv.__init__(
            self,
            get_asset_xml("twod_maze.xml"),
            2,
            xbounds=[-0.3, 0.3],
            ybounds=[-0.3, 0.3],
        )

    def step(self, a):
        self.do_simulation(a, self.frame_skip)
        ob = self._get_obs()
        pos = ob[0:2]
        dist = np.sum(np.abs(pos - TARGET))  # np.linalg.norm(pos - TARGET)
        reward = -(dist)

        reward_ctrl = -np.square(a).sum()
        reward += 1e-3 * reward_ctrl

        if self.verbose:
            print(pos, reward)
        self.episode_length += 1
        done = self.episode_length >= self.max_episode_length
        return ob, reward, done, {"distance": dist}

    def reset_model(self):
        self.episode_length = 0
        qpos = self.init_qpos + self.np_random.uniform(
            size=self.model.nq, low=-0.01, high=0.01
        )
        qvel = self.init_qvel + self.np_random.uniform(
            size=self.model.nv, low=-0.01, high=0.01
        )
        self.set_state(qpos, qvel)
        return self._get_obs()

    def _get_obs(self):
        # return np.concatenate([self.sim.data.qpos, self.sim.data.qvel]).ravel()
        return np.concatenate([self.sim.data.qpos]).ravel() - INIT_POS

    def viewer_setup(self):
        pass



from typing import List, NamedTuple, Optional

import numpy as np


class Trajectory(NamedTuple):
    """A trajectory, e.g. a one episode rollout from an expert policy."""

    acts: np.ndarray
    """Actions, shape (trajectory_len, ) + action_shape."""

    obs: np.ndarray
    """Observations, shape (trajectory_len + 1, ) + observation_shape."""

    rews: np.ndarray
    """Reward, shape (trajectory_len, )."""

    infos: Optional[List[dict]]
    """A list of info dicts, length trajectory_len."""



"""Constructs deep network reward models."""

from abc import ABC, abstractmethod
from typing import Iterable, Optional, Sequence, Tuple

import gym
import tensorflow as tf

from imitation.util import networks, serialize


class RewardNet(serialize.Serializable, ABC):
    """Abstract reward network.

    This class assumes that the caller will set the default TensorFlow Session
    and initialize the network's variables.

    Attributes:
      observation_space: The observation space of `obs_ph` and `next_obs_ph`.
      action_space: The action space of `act_ph`.
      obs_ph (tf.Tensor): previous observation placeholder.
      act_ph (tf.Tensor): action placeholder.
      next_obs_ph (tf.Tensor): next observation placeholder.
      _params (dict): parameters to serialize in `save`, used as keyword
          arguments for constructor by `load`.
      _layers (dict): CheckpointableBase objects, e.g. a TensorFlow layer,
          saved by `save` and restored by `load`.
    """

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        *,
        scale: bool = False,
        use_state: bool = True,
        use_action: bool = True,
        use_next_state: bool = False,
        use_done: bool = False,
    ):
        """Builds a reward network.

        Args:
            observation_space: The observation space.
            action_space: The action space.
            scale: Whether to scale the input.
            use_state: Whether state is included in inputs to network.
            use_action: Whether action is included in inputs to network.
            use_next_state: Whether next state is included in inputs to network.
            use_done: Whether episode termination is included in inputs to network.
        """

        self.observation_space = observation_space
        self.action_space = action_space
        self.scale = scale

        phs, inps = networks.build_inputs(observation_space, action_space, scale)
        self.obs_ph, self.act_ph, self.next_obs_ph, self.done_ph = phs
        self.obs_inp, self.act_inp, self.next_obs_inp, self.done_inp = inps

        inputs = []
        inputs += [self.obs_inp] if use_state else []
        inputs += [self.act_inp] if use_action else []
        inputs += [self.next_obs_inp] if use_next_state else []
        if len(inputs) == 0:
            raise ValueError(
                "At least one of `use_state`, `use_action` and `use_next_state` "
                "must be true."
            )
        inputs += [self.done_inp] if use_done else []

        with tf.variable_scope("theta_network"):
            self._theta_output, theta_layers = self.build_theta_network(inputs)

        self._layers = theta_layers

    @property
    @abstractmethod
    def reward_output_train(self):
        """A Tensor holding the training reward associated with each timestep.

        Different concrete subclasses will require different placeholders to be
        filled to calculate this output, but for all subclasses, filling the
        following placeholders will be sufficient:

        ```
        self.obs_ph
        self.act_ph
        self.next_obs_ph
        ```

        Returns:
          tf.Tensor: A (None,) shaped Tensor holding
              the training reward associated with each timestep.
        """
        pass

    @property
    def reward_output_test(self):
        """A Tensor holding the test reward associated with each timestep.

        Note this is the reward we use for transfer learning.

        Different concrete subclasses will require different
        placeholders to be filled to calculate this output, but for all
        subclasses, filling the following placeholders will be sufficient:

        ```
        self.obs_ph
        self.act_ph
        self.next_obs_ph
        ```

        Returns:
          tf.Tensor: A (None,) shaped Tensor holding
            the test reward associated with each timestep.
        """
        return self._theta_output

    @abstractmethod
    def build_theta_network(
        self, inputs: Sequence[tf.Tensor]
    ) -> Tuple[tf.Tensor, networks.LayersDict]:
        """Builds the test reward network.

        The output of the network is the same as the reward used for transfer
        learning, and is the Tensor returned by `self.reward_output_test()`.

        Args:
          obs_input: The observation input. Its shape is
              `((None,) + self.env.observation_space.shape)`.
          act_input: The action input. Its shape is
              `((None,) + self.env.action_space.shape)`. The None dimension is
              expected to be the same as None dimension from `obs_input`.

        Returns:
          A tuple (theta_output, layers) where
            * theta_output is a reward prediction for each of the inputs,
              of shape `(None,)`;
            * layers is a dictionary mapping to individual checkpointable layers.
        """
        pass

    def build_summaries(self):
        tf.summary.histogram("train_reward", self.reward_output_train)
        tf.summary.histogram("test_reward", self.reward_output_test)


class RewardNetShaped(RewardNet):
    """Abstract reward network with a phi network to shape training reward.

    This RewardNet formulation matches Equation (4) in the AIRL paper.
    Note that the experiments in Table 2 of the same paper showed shaped
    training rewards to be inferior to an unshaped training rewards in
    a Pendulum environment imitation learning task (and maybe HalfCheetah).
    (See original implementation of Pendulum experiment's reward function at
    https://github.com/justinjfu/inverse_rl/blob/master/inverse_rl/models/imitation_learning.py#L374)

    To make a concrete subclass, implement `build_phi_network()` and
    `build_theta_network()`.
    """

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        *,
        discount_factor: float = 0.99,
        **kwargs,
    ):
        super().__init__(observation_space, action_space, **kwargs)
        self._discount_factor = discount_factor

        with tf.variable_scope("phi_network"):
            res = self.build_phi_network(self.obs_inp, self.next_obs_inp)
            # end_potential is the potential when the episode terminates.
            if discount_factor == 1.0:
                # If undiscounted, terminal state must have potential 0.
                end_potential = tf.constant(0.0)
            else:
                # Otherwise, it can be arbitrary, so make a trainable variable.
                end_potential = tf.Variable(
                    name="end_phi", shape=(), dtype=tf.float32, initial_value=0.0
                )
                self._layers.update(end_potential=end_potential)
            self._old_shaping_output, self._new_shaping_output, phi_layers = res
            self._layers.update(**phi_layers)

        with tf.variable_scope("f_network"):
            new_shaping = (
                self.done_inp * end_potential
                + (1 - self.done_inp) * self._new_shaping_output
            )
            self._shaped_reward_output = (
                self._theta_output
                + self._discount_factor * new_shaping
                - self._old_shaping_output
            )

    @property
    def reward_output_train(self):
        """A Tensor holding the (shaped) training reward of each timestep.

        Requires the following placeholders to be filled:

        ```
        self.obs_ph
        self.act_ph
        self.next_obs_ph
        ```

        Returns:
          tf.Tensor: A (None,) shaped Tensor holding
              the training reward associated with each timestep.
        """
        return self._shaped_reward_output

    @abstractmethod
    def build_phi_network(
        self,
        obs_input: tf.Tensor,
        next_obs_input: tf.Tensor,
    ) -> Tuple[tf.Tensor, tf.Tensor, networks.LayersDict]:
        """Build the reward shaping network (disentangles dynamics from reward).

        XXX: We could potentially make it easier on the subclasser by requiring
        only one input. ie build_phi_network(obs_input). Later in
        _build_f_network, I could stack Tensors of old and new observations,
        pass them simulatenously through the network, and then unstack the
        outputs. Another way to do this would be to pass in a single
        rank 3 obs_input with shape `(2, None) + self.env.observation_space`.

        Args:
          obs_input: The old observations (corresponding to the state at which
              the current action is made). The shape of this Tensor should be
              `(None,) + self.env.observation_space.shape`.
          next_obs_input: The new observations (corresponding to the state that we
              transition to after this state-action pair.

        Returns:
          A tuple (old_shaping_output, new_shaping_output, layers) where
            * old_shaping_output is a reward shaping prediction for each of the old
              observation inputs, with shape `(None,)`.
            * new_shaping_output is a reward shaping prediction for each of the new
              observation inputs, with shape `(None,)`.
            * layers is a dictionary mapping to individual checkpointable layers.
        """
        pass

    def build_summaries(self):
        super().build_summaries()
        tf.summary.histogram("shaping_old", self._old_shaping_output)
        tf.summary.histogram("shaping_new", self._new_shaping_output)


class BasicRewardNet(RewardNet, serialize.LayersSerializable):
    """An unshaped reward net with simple, default settings.

    Intended to match the reward network trained for the original AIRL pendulum
    experiments. Right now it has a linear function approximator for the theta network,
    not sure if this is what I want.
    """

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        *,
        theta_units: Optional[Iterable[int]] = None,
        theta_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """Builds a simple reward network.

        Args:
          observation_space: The observation space.
          action_space: The action space.
          theta_units: Number of hidden units at each layer of the feedforward
              reward network theta.
          theta_kwargs: Arguments passed to `build_basic_theta_network`.
          kwargs: Passed through to RewardNet.
        """
        params = locals()
        del params["kwargs"]
        params.update(kwargs)

        self.theta_units = theta_units
        self.theta_kwargs = theta_kwargs or {}
        RewardNet.__init__(self, observation_space, action_space, **kwargs)
        serialize.LayersSerializable.__init__(**params, layers=self._layers)

    def build_theta_network(self, inputs: Sequence[tf.Tensor]):
        return networks.build_and_apply_mlp(
            inputs, self.theta_units, **self.theta_kwargs
        )

    @property
    def reward_output_train(self):
        """Training reward is the same as the test reward, since no shaping."""
        return self.reward_output_test


def build_basic_phi_network(
    hid_sizes: Optional[Iterable[int]],
    obs_input: tf.Tensor,
    next_obs_input: tf.Tensor,
    **kwargs: dict,
):
    """Builds a potential network depending on specified observation.

    Arguments:
      hid_sizes: Number of units at each hidden layer. Default is (32, 32).
      obs_input: Previous observation.
      next_obs_input: Next observation.
      **kwargs: Passed through to `util.build_mlp`.

    Returns:
      Tuple[tf.Tensor, tf.Tensor]: potential for the old and new observations.
    """
    if hid_sizes is None:
        hid_sizes = (32, 32)

    with tf.variable_scope("phi", reuse=tf.AUTO_REUSE):
        old_o = tf.layers.flatten(obs_input)
        new_o = tf.layers.flatten(next_obs_input)

        # Weight share, just with different inputs old_o and new_o
        phi_mlp = networks.build_mlp(hid_sizes=hid_sizes, name="shaping", **kwargs)
        old_shaping_output = networks.sequential(old_o, phi_mlp)
        new_shaping_output = networks.sequential(new_o, phi_mlp)

    return old_shaping_output, new_shaping_output, phi_mlp


class BasicShapedRewardNet(RewardNetShaped, serialize.LayersSerializable):
    """A shaped reward network with simple, default settings.

    With default parameters this RewardNet has two hidden layers [32, 32]
    for the theta network and reward shaping phi network.

    This network is feed-forward and flattens inputs, so is a poor choice for
    training on pixel observations.
    """

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        *,
        theta_units: Optional[Iterable[int]] = None,
        theta_kwargs: Optional[dict] = None,
        phi_units: Optional[Iterable[int]] = None,
        phi_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """Builds a simple shaped reward network.

        Args:
          observation_space: The observation space.
          action_space: The action space.
          theta_units: Number of hidden units at each layer of the feedforward
              reward network theta.
          theta_kwargs: Arguments passed to `build_basic_theta_network`.
          phi_units: Number of hidden units at each layer of the feedforward
              potential network phi.
          phi_kwargs: Arguments passed to `build_basic_phi_network`.
          kwargs: Passed through to `RewardNetShaped`.
        """
        params = locals()
        del params["kwargs"]
        params.update(kwargs)

        self.theta_units = theta_units
        self.phi_units = phi_units
        self.theta_kwargs = theta_kwargs or {}
        self.phi_kwargs = phi_kwargs or {}
        RewardNetShaped.__init__(
            self,
            observation_space,
            action_space,
            **kwargs,
        )
        serialize.LayersSerializable.__init__(**params, layers=self._layers)

    def build_theta_network(self, inputs):
        return networks.build_and_apply_mlp(
            inputs, self.theta_units, **self.theta_kwargs
        )

    def build_phi_network(self, obs_input, next_obs_input):
        return build_basic_phi_network(
            self.phi_units, obs_input, next_obs_input, **self.phi_kwargs
        )



"""Utilities and definitions shared by reward-related code."""

import functools
from typing import Callable

import numpy as np
from stable_baselines.common import vec_env

RewardFn = Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray]


def _reward_fn_normalize_inputs(
    obs: np.ndarray,
    acts: np.ndarray,
    next_obs: np.ndarray,
    dones: np.ndarray,
    *,
    reward_fn: RewardFn,
    vec_normalize: vec_env.VecNormalize,
    norm_reward: bool = True,
) -> np.ndarray:
    """Combine with `functools.partial` to create an input-normalizing RewardFn.

    Args:
        reward_fn: The reward function that normalized inputs are evaluated on.
        vec_normalize: Instance of VecNormalize used to normalize inputs and
            rewards.
        norm_reward: If True, then also normalize reward before returning.

    Returns:
        The possibly normalized reward.
    """
    norm_obs = vec_normalize.normalize_obs(obs)
    norm_next_obs = vec_normalize.normalize_obs(next_obs)
    rew = reward_fn(norm_obs, acts, norm_next_obs, dones)
    if norm_reward:
        rew = vec_normalize.normalize_reward(rew)
    return rew


def build_norm_reward_fn(*, reward_fn, vec_normalize, **kwargs) -> RewardFn:
    """Reward function that automatically normalizes inputs.

    See _reward_fn_normalize_inputs for argument documentation.
    """
    return functools.partial(
        _reward_fn_normalize_inputs,
        reward_fn=reward_fn,
        vec_normalize=vec_normalize,
        **kwargs,
    )



import datetime
import functools
import os
import uuid
from typing import Optional, Type, Union

import gym
import numpy as np
import stable_baselines
from gym.wrappers import TimeLimit
from stable_baselines import bench
from stable_baselines.common.base_class import BaseRLModel
from stable_baselines.common.policies import BasePolicy, MlpPolicy
from stable_baselines.common.vec_env import DummyVecEnv, SubprocVecEnv, VecEnv

from imitation.data import wrappers


def make_unique_timestamp() -> str:
    """Timestamp, with random uuid added to avoid collisions."""
    ISO_TIMESTAMP = "%Y%m%d_%H%M%S"
    timestamp = datetime.datetime.now().strftime(ISO_TIMESTAMP)
    random_uuid = uuid.uuid4().hex[:6]
    return f"{timestamp}_{random_uuid}"


def make_vec_env(
    env_name: str,
    n_envs: int = 8,
    seed: int = 0,
    parallel: bool = False,
    log_dir: Optional[str] = None,
    max_episode_steps: Optional[int] = None,
) -> VecEnv:
    """Returns a VecEnv initialized with `n_envs` Envs.

    Args:
        env_name: The Env's string id in Gym.
        n_envs: The number of duplicate environments.
        seed: The environment seed.
        parallel: If True, uses SubprocVecEnv; otherwise, DummyVecEnv.
        log_dir: If specified, saves Monitor output to this directory.
        max_episode_steps: If specified, wraps each env in a TimeLimit wrapper
            with this episode length. If not specified and `max_episode_steps`
            exists for this `env_name` in the Gym registry, uses the registry
            `max_episode_steps` for every TimeLimit wrapper (this automatic
            wrapper is the default behavior when calling `gym.make`). Otherwise
            the environments are passed into the VecEnv unwrapped.
    """
    # Resolve the spec outside of the subprocess first, so that it is available to
    # subprocesses running `make_env` via automatic pickling.
    spec = gym.spec(env_name)

    def make_env(i, this_seed):
        # Previously, we directly called `gym.make(env_name)`, but running
        # `imitation.scripts.train_adversarial` within `imitation.scripts.parallel`
        # created a weird interaction between Gym and Ray -- `gym.make` would fail
        # inside this function for any of our custom environment unless those
        # environments were also `gym.register()`ed inside `make_env`. Even
        # registering the custom environment in the scope of `make_vec_env` didn't
        # work. For more discussion and hypotheses on this issue see PR #160:
        # https://github.com/HumanCompatibleAI/imitation/pull/160.
        env = spec.make()

        # Seed each environment with a different, non-sequential seed for diversity
        # (even if caller is passing us sequentially-assigned base seeds). int() is
        # necessary to work around gym bug where it chokes on numpy int64s.
        env.seed(int(this_seed))

        if max_episode_steps is not None:
            env = TimeLimit(env, max_episode_steps)
        elif spec.max_episode_steps is not None:
            env = TimeLimit(env, max_episode_steps=spec.max_episode_steps)

        # Use Monitor to record statistics needed for Baselines algorithms logging
        # Optionally, save to disk
        log_path = None
        if log_dir is not None:
            log_subdir = os.path.join(log_dir, "monitor")
            os.makedirs(log_subdir, exist_ok=True)
            log_path = os.path.join(log_subdir, f"mon{i:03d}")

        env = bench.Monitor(env, log_path)
        env = wrappers.RolloutInfoWrapper(env)
        return env

    rng = np.random.RandomState(seed)
    env_seeds = rng.randint(0, (1 << 31) - 1, (n_envs,))
    env_fns = [functools.partial(make_env, i, s) for i, s in enumerate(env_seeds)]
    if parallel:
        # See GH hill-a/stable-baselines issue #217
        return SubprocVecEnv(env_fns, start_method="forkserver")
    else:
        return DummyVecEnv(env_fns)


def init_rl(
    env: Union[gym.Env, VecEnv],
    model_class: Type[BaseRLModel] = stable_baselines.PPO2,
    policy_class: Type[BasePolicy] = MlpPolicy,
    **model_kwargs,
):
    """Instantiates a policy for the provided environment.

    Args:
        env: The (vector) environment.
        model_class: A Stable Baselines RL algorithm.
        policy_class: A Stable Baselines compatible policy network class.
        model_kwargs (dict): kwargs passed through to the algorithm.
          Note: anything specified in `policy_kwargs` is passed through by the
          algorithm to the policy network.

    Returns:
      An RL algorithm.
    """
    return model_class(
        policy_class, env, **model_kwargs
    )  # pytype: disable=not-instantiable


def docstring_parameter(*args, **kwargs):
    """Treats the docstring as a format string, substituting in the arguments."""

    def helper(obj):
        obj.__doc__ = obj.__doc__.format(*args, **kwargs)
        return obj

    return helper



"""Heatmaps and reward plotting code for debugging MountainCar."""

import pathlib
import pickle
from typing import Dict, List, Optional, Sequence, Union

import gym
import numpy as np
from matplotlib import pyplot as plt
from stable_baselines.common import vec_env

from imitation.data import rollout, types
from imitation.policies import serialize as policies_serialize
from imitation.rewards import common
from imitation.rewards import serialize as rewards_serialize

MC_POS_MIN, MC_POS_MAX = -1.2, 0.6
MC_VEL_MIN, MC_VEL_MAX = -0.07, 0.07
MC_GOAL_POS = 0.5
MC_NUM_ACTS = 3

ACT_NAMES = ["left", "neutral", "right"]


def _make_next_mc_obs(obs, acts) -> np.ndarray:
    """Utility for calculating the MountainCar-v0 next observation s'.

    Required for evaluating AIRL reward at arbitrary (s, a) points.
    """
    env = gym.make("MountainCar-v0")
    obs = np.array(obs)
    acts = np.array(acts)
    next_obs = []
    for ob, act in zip(obs, acts):
        assert ob.shape == (2,)
        env.reset()
        env.unwrapped.state = ob
        next_ob = env.step(act)[0]
        next_obs.append(next_ob)
    return np.array(next_obs)


def make_heatmap(
    act: int,
    reward_fn: common.RewardFn,
    n_pos_step: int = 18,
    n_vel_step: int = 14,
    mark_goal: bool = True,
    gen_trajs: Optional[List[types.Trajectory]] = None,
    exp_trajs: Optional[List[types.Trajectory]] = None,
    legend_on: bool = True,
    title: bool = None,
    heatmap: bool = True,
    filter_trans_by_act: bool = True,
) -> plt.Figure:
    """Make a MountainCar heatmap of rewards for a particular action.

    X axis is position. Y axis is velocity.

    Args:
        act: The MountainCar action number whose reward we are evaluating.
            Should be 0, 1, or 2.
        reward_fn: Reward function. Should accept unnormalized inputs.
        n_pos_step: The number of squares that the x axis of the heatmap is divided
            into.
        n_vel_step: The number of squares that the y axis of the heatmap is divided
            into.
        gen_trajs: A list of generator trajectories to
            scatterplot on top of the heatmap.
        exp_trajs: A list of exp trajectories to scatterplot on
            top of the heatmap.
        legend_on: Whether to plot the legend.
        title: Custom title.
        heatmap: Whether to plot the heatmap.
        filter_trans_by_act: If True, then filter out transitions from
            `gen_trajs` and `exp_trajs` that don't use action `act` before
            scatterplotting.
    """
    assert 0 <= act < MC_NUM_ACTS

    def convert_traj_to_coords_filtered(trajs: Sequence[types.Trajectory]):
        trans = rollout.flatten_trajectories(trajs)
        obs = trans.obs
        if filter_trans_by_act:
            obs = obs[trans.acts == act]
        return obs[:, 0], obs[:, 1]

    fig, ax = plt.subplots()
    if heatmap:
        pos_space = np.linspace(MC_POS_MIN, MC_POS_MAX, n_pos_step, endpoint=True)
        vel_space = np.linspace(MC_VEL_MIN, MC_VEL_MAX, n_vel_step, endpoint=True)
        obs_vec = np.array([[p, v] for p in pos_space for v in vel_space])
        acts_vec = np.array([act] * len(obs_vec))
        next_obs_vec = _make_next_mc_obs(obs_vec, acts_vec)
        dones = np.zeros(len(acts_vec), dtype=np.bool)

        rew = reward_fn(obs_vec, acts_vec, next_obs_vec, dones)
        # Transpose because `pcolor` (confusingly) expects its first two arguments
        # to be XY, but its matrix argument to be in RC format.
        rew_matrix = rew.reshape(n_pos_step, n_vel_step).T
        c = ax.pcolor(pos_space, vel_space, rew_matrix)
        fig.colorbar(c, ax=ax)

    if mark_goal:
        ax.axvline(
            x=MC_GOAL_POS, linestyle="--", label=f"goal state (pos={MC_GOAL_POS})"
        )
    if exp_trajs is not None:
        X, Y = convert_traj_to_coords_filtered(exp_trajs)
        ax.scatter(X, Y, marker="o", label="expert samples", alpha=0.2)
    if gen_trajs is not None:
        X, Y = convert_traj_to_coords_filtered(gen_trajs)
        ax.scatter(X, Y, marker="o", c="yellow", label="policy samples", alpha=0.2)

    if title is None:
        title = f"Action {ACT_NAMES[act]}"
    ax.set_title(title)
    if legend_on:
        ax.legend(loc="center left", bbox_to_anchor=(0, 1.3))

    return fig


def batch_reward_heatmaps(
    checkpoints_dir: Union[str, pathlib.Path],
    n_gen_trajs: int = 50,
    exp_trajs: Optional[List[types.Trajectory]] = None,
) -> Dict[pathlib.Path, plt.Figure]:
    """Build multiple mountain car reward heatmaps from a checkpoint directory.

    One plot is generated for every combination of action and checkpoint timestep.

    Args:
        checkpoints_dir: Path to `checkpoint/` directory from AIRL or GAIL output
            directory.
        n_gen_trajs: The number of trajectories to rollout using each generator
            checkpoint. The transitions in the trajectory are scatterplotted on top of
            the heatmap from the same checkpoint timestamp. Nonpositive indicates that
            no trajectories should be plotted.
        exp_trajs: Expert trajectories for scatterplotting. Generator trajectories
            are dynamically generated from generator checkpoints.

    Returns:
        A dictionary mapping relative paths to `plt.Figure`. Every key is of the
        form "{action_name}/{checkpoint_step}" where action_name is "left",
        "neutral", or "right".
    """
    result = {}
    venv = vec_env.DummyVecEnv([lambda: gym.make("MountainCar-v0")])
    checkpoints_dir = pathlib.Path(checkpoints_dir)
    for checkpoint_dir in sorted(checkpoints_dir.iterdir()):
        vec_normalize_path = checkpoint_dir / "gen_policy" / "vec_normalize.pkl"
        discrim_path = checkpoint_dir / "discrim"
        policy_path = checkpoint_dir / "gen_policy"

        if n_gen_trajs > 0:
            # `load_policy` automatically loads VecNormalize for policy evaluation.
            with policies_serialize.load_policy(
                "ppo2", str(policy_path), venv
            ) as gen_policy:
                gen_trajs = rollout.generate_trajectories(
                    gen_policy, venv, sample_until=rollout.min_episodes(n_gen_trajs)
                )
        else:
            gen_trajs = None

        # `gen_trajs` contains unnormalized observations.
        # Load VecNormalize for use in RewardFn, which doesn't automatically
        # normalize input observations.
        with open(vec_normalize_path, "rb") as f:
            vec_normalize = pickle.load(f)  # type: vec_env.VecNormalize
        vec_normalize.training = False

        reward_fn_ctx = rewards_serialize.load_reward("DiscrimNet", discrim_path, venv)
        with reward_fn_ctx as reward_fn:
            norm_rew_fn = common.build_norm_reward_fn(
                reward_fn=reward_fn, vec_normalize=vec_normalize
            )
            for act in range(MC_NUM_ACTS):
                fig = make_heatmap(
                    act=act,
                    reward_fn=norm_rew_fn,
                    gen_trajs=gen_trajs,
                    exp_trajs=exp_trajs,
                )
                path = pathlib.Path(ACT_NAMES[act], checkpoint_dir.name)
                result[path] = fig
    return result


def plot_reward_vs_time(
    trajs_dict: Dict[str, List[types.Trajectory]],
    reward_fn: common.RewardFn,
    preferred_colors: Optional[Dict[str, str]] = None,
) -> plt.Figure:
    """Plots a reward versus timestep line for each Trajectory.

    Args:
        trajs_dict: A dictionary mapping rollout labels (e.g. "expert" or
            "gen policy") to rollouts associated with those labels.
        reward_fn: Reward function for evaluating rollout rewards.
        preferred_colors: An optional dictionary mapping rollout labels to
            preferred line colors.

    Returns:
        The figure.
    """
    if preferred_colors is None:
        preferred_colors = {}
    fig, ax = plt.subplots()

    for i, (trajs_label, trajs_list) in enumerate(trajs_dict.items()):
        X = []
        Y = []
        for traj in trajs_list:
            T = len(traj.rews)
            X.extend(range(T))
            dones = np.zeros(T, dtype=np.bool)
            dones[-1] = True
            rews = reward_fn(traj.obs[:-1], traj.acts, traj.obs[1:], dones)
            Y.extend(rews)
        color = preferred_colors.get(trajs_label, None)
        ax.plot(X, Y, label=trajs_label, c=color)
    ax.set_xlabel("timestep")
    ax.set_ylabel("test reward")
    ax.legend()
    return fig
